open Util
open Sequence

type altype = Local | Global | GlobalEF

type trace_dir = Match | XGapOpen | YGapOpen| XGapExtend | YGapExtend | Stop;;

let gap_open = ref 3.0			(** Gap open *)
let gap_ext  = ref 1.5			(** Gap extend *)
let altype = ref Local			(** Local, Global, or GlobalES alignment *)
let zscore = ref false			(** Whether or not to calculate z-scores *)
let nshuffles = ref 100			(** How many shuffles to use for z-score computation *)

(** This is the generic, low-level alignment routine. It fills the
    traceback [tr] and populates [xmap] and [ymap] with the best
    alignment according to the specifications. *)
(* Remember: first row/column of e f and v have to do with initial
   conditions so they are each one row/col bigger than sequence
   length. scores is same size, index by taking 1 off of the i,j
   indexes.  *)
(* Conceptually first sequence X is layed out along the rows, second
   sequence Y along the columns. E corresponds to gaps in X (consume
   characters of Y), and F to gaps in Y. For the initial conditions
   in E and F, the opposite is true as the 0th column corresponds to
   all gaps in Y and the 0th row corresponds to all gaps in X. *)
let gen_align v e f tr xmap ymap scores m n =
  assert(check_matsize (m+1) (n+1) v);
  assert(check_matsize (m+1) (n+1) e);
  assert(check_matsize (m+1) (n+1) f);
  assert(check_matsize m n scores);

  let altype,go,ge = !altype, !gap_open, !gap_ext in
  (* Set up the scoring according to the local/global convention *)
  fill_matrix v 0.0; fill_matrix e 0.0; fill_matrix f 0.0;  
  begin match altype with
    Local -> fill_matrix tr (Stop,Stop,Stop); 
  | Global ->
      tr.(0).(0) <- (Stop,Stop,Stop);
      e.(1).(0) <- -. (go +. ge); tr.(1).(0) <- (YGapOpen,YGapOpen,YGapOpen);
      f.(0).(1) <- -. (go +. ge); tr.(0).(1) <- (XGapOpen,XGapOpen,XGapOpen);
      v.(1).(0) <- e.(1).(0); v.(0).(1) <- f.(0).(1);
      for i=2 to m do
	e.(i).(0) <- e.(i-1).(0) -. ge;
	v.(i).(0) <- e.(i).(0);
	tr.(i).(0) <- (YGapExtend,YGapExtend,YGapExtend);
      done;
      for i=2 to n do
	f.(0).(i) <- f.(0).(i-1) -. ge;
	v.(0).(i) <- f.(0).(i);
	tr.(0).(i) <- (XGapExtend,XGapExtend,XGapExtend);
      done;
  | GlobalEF ->
      tr.(0).(0) <- (Stop,Stop,Stop);
      e.(1).(0) <- 0.0; tr.(1).(0) <- (YGapOpen,YGapOpen,YGapOpen);
      f.(0).(1) <- 0.0; tr.(0).(1) <- (XGapOpen,XGapOpen,XGapOpen);
      v.(1).(0) <- e.(1).(0); v.(0).(1) <- f.(0).(1);
      for i=2 to m do
	e.(i).(0) <- 0.0;
	v.(i).(0) <- e.(i).(0);
	tr.(i).(0) <- (YGapExtend,YGapExtend,YGapExtend);
      done;
      for i=2 to n do
	f.(0).(i) <- 0.0;
	v.(0).(i) <- f.(0).(i);
	tr.(0).(i) <- (XGapExtend,XGapExtend,XGapExtend);
      done;
      
  end;

  let choose (s1:float) (s2:float) (mv1:trace_dir) (mv2:trace_dir) =
    if s1 >= s2 then s1,mv1 else s2,mv2 in

  (* Fill in the DP matrices *)
  let best,ibest,jbest = ref 0.0, ref 0, ref 0 in
  for i=1 to m do
    for j=1 to n do
      let es,emove = choose (e.(i).(j-1)-.ge) (v.(i).(j-1)-.go-.ge) XGapExtend XGapOpen in
      let fs,fmove = choose (f.(i-1).(j)-.ge) (v.(i-1).(j)-.go-.ge) YGapExtend YGapOpen in
      let ef,efmove= choose fs es fmove emove in
      let ms = v.(i-1).(j-1) +. scores.(i-1).(j-1) in
      let vs,vmove = choose ms ef Match efmove in
      let vs,vmove = match altype with
	  Local -> choose 0.0 vs Stop vmove
	| Global | GlobalEF -> vs,vmove in
      e.(i).(j) <- es;
      f.(i).(j) <- fs;
      v.(i).(j) <- vs;
      tr.(i).(j) <- (vmove,emove,fmove);
      if !best < v.(i).(j) then begin
	best := v.(i).(j);
	ibest := i;
	jbest := j;
      end;
    done;
  done;

  let score,iend,jend = match altype with 
      Local-> !best,!ibest,!jbest 
    | Global-> v.(m).(n),m,n 
    | GlobalEF->			
	(* Hackery: find the best starting point for the alignment
	   along last row or last column, adjust path from m,n to
	   lead the alignment there, keep best score intact *)
	best:=v.(m).(n); ibest:=m; jbest:=n;
	for i=m-1 downto 0 do
	  if v.(i).(n) >= !best then 
	    (best := v.(i).(n); ibest:=i; jbest:=n);
	done;
	for j=n-1 downto 0 do
	  if v.(m).(j) >= !best then 
	    (best := v.(m).(j); ibest:=m; jbest:=j);
	done;
	if !ibest < m then begin
	  for i=m downto !ibest+2 do
	    tr.(i).(n) <- YGapExtend,YGapExtend,YGapExtend
	  done;
	  tr.(!ibest+1).(n) <- YGapOpen,YGapOpen,YGapOpen;
	end
	else if !jbest < n then begin
	  for j=n downto !jbest+2 do
	    tr.(m).(j) <- XGapExtend,XGapExtend,XGapExtend
	  done;
	  tr.(m).(!jbest+1) <- XGapOpen,XGapOpen,XGapOpen;
	end;
	!best,m,n;
  in
  
  (* Traceback procedure, recursively step through starting at i j,
     perform [op] at each step. *)
  let rec traceback op st i j =
    let v,e,f = tr.(i).(j) in
    let mv = match st with
	`V -> v | `E -> e | `F -> f in
    op i j mv;
    let st2,i2,j2 = match mv with
	Match -> `V, (i-1), (j-1)
      | XGapOpen -> `V, i, (j-1)
      | YGapOpen -> `V, (i-1), j
      | XGapExtend -> `E, i, (j-1)
      | YGapExtend -> `F, (i-1), j
      | Stop -> `V, -1, -1 in
    if mv <> Stop then traceback op st2 i2 j2
  in
  (* Calculate size of alignment *)
  let alsize =
    let n = ref 0 in
    traceback (fun _ _ _ -> incr n) `V iend jend;
    !n-1 in
  assert(Array.length xmap >= alsize && Array.length ymap >= alsize);
  let ii = ref (alsize-1) in
  (* Fill maps with proper indices of their characters *)
  let fillmap i j dir = 
    begin match dir with
      Match -> xmap.(!ii)<-Some(i-1); ymap.(!ii)<-Some(j-1); 
    | XGapOpen | XGapExtend -> 
	xmap.(!ii) <- None; ymap.(!ii) <- Some(j-1); 
    | YGapOpen | YGapExtend -> 
	xmap.(!ii) <- Some(i-1); ymap.(!ii) <- None;
    | Stop -> () end;
    decr ii; in
  traceback fillmap `V iend jend;
  score,alsize;;
(*   let s = ref 0.0 in *)
(*   let vscore i j dir = match dir with *)
(*       Match -> s := !s +. scores.(i-1).(j-1) *)
(*     | XGapOpen | YGapOpen -> s := !s -. go -. ge *)
(*     | XGapExtend| YGapExtend -> s := !s -. ge *)
(*     | Stop -> () in *)
(*   traceback vscore `V iend jend; *)


(** Compute the average and standard deviation of random shuffles of
    two sequences.  Note that this alters the scoring matrix by
    permuting columns. *)
let avgstd_align v e f tr xmap ymap scores m n =
  let avgstd = Avgstd.make () in
  let nshuffles = !nshuffles in
  for i=1 to nshuffles do
    Util.permute 3 scores;
    let score,_ = gen_align v e f tr xmap ymap scores m n in
    Avgstd.update avgstd score;
  done;
  let a,s,_ = Avgstd.finalize avgstd in
  a,s

(** Wrapper for gen_align which will execute z-score calcuation if
    required at the module level. *)
let align_seqs v e f tr xmap ymap scores m n =
  let avg,std =
    if !zscore then 			(* Get z-scores if needed *)
      let scores2 = Array.copy scores in
      avgstd_align v e f tr xmap ymap scores2 m n
    else				(* O/w use nonmodifying avg,std *)
      0.0,1.0
  in
  let score,size = gen_align v e f tr xmap ymap scores m n in
  let score = (score -. avg) /. std in
  score,size


(** Allocate space for two sequences of length [m] and [n], apply
    [scorefuncs] on the scoring matrix and align them. Return the
    score, length, maps for the alignment, and the DP matrices *)
let allocate_align x y =
  let m,n = x.seqlen, y.seqlen in
  let mmat = Array.make_matrix in
  let scores = mmat m n 0.0 in
  let v = mmat (m+1) (n+1) 0.0 in
  let e = mmat (m+1) (n+1) 0.0 in
  let f = mmat (m+1) (n+1) 0.0 in
  let tr= mmat (m+1) (n+1) (Stop,Stop,Stop) in
  let xmap = Array.make (m+n) None in
  let ymap = Array.make (m+n) None in
  (* Set up the scoring functions *)
  Scoring.setup_scoring_mat ~reset:false scores x y;
  let score,size = align_seqs v e f tr xmap ymap scores m n in
  score,size,xmap,ymap,(v,e,f,tr,scores);;

(** Allocate a matrices to do alignments between sequences of maximum
    size [n] *)
let allocate_max n =
  let m = n in
  let mmat = Array.make_matrix in
  let scores = mmat m n 0.0 in
  let v = mmat (m+1) (n+1) 0.0 in
  let e = mmat (m+1) (n+1) 0.0 in
  let f = mmat (m+1) (n+1) 0.0 in
  let tr= mmat (m+1) (n+1) (Stop,Stop,Stop) in
  let xmap = Array.make (m+n) None in
  let ymap = Array.make (m+n) None in
  scores,v,e,f,tr,xmap,ymap;;


(** Thrown when two gap characters are found aligned, returns the
    alignment position where occurs *)
exception Aligned_gaps of int;;

(** Score an alignment *)
let score_alignment xmap ymap scores start stop =
  let go,ge = !gap_open, !gap_ext in

  let rec loop i gapx gapy s =
    if not (i<stop) then s
    else match gapx,gapy,xmap.(i),ymap.(i) with
      _,_,Some(xi),Some(yi) ->		(* Match *)
	loop (i+1) false false (s +. scores.(xi).(yi))
    | false,false,None,Some(_)
    | false,false,Some(_),None
    | true,false,Some(_),None
    | false,true,None,Some(_) ->	(* Gap opening *)
	loop (i+1) (xmap.(i)=None) (ymap.(i)=None) (s -. go -. ge)
    | true,false,None,Some(_)
    | false,true,Some(_),None ->	(* Gap extension *)
	loop (i+1) (xmap.(i)=None) (ymap.(i)=None) (s -. ge)
    | _ -> raise (Aligned_gaps i)
  in loop start false false 0.0;;

(** Remove endspace from an alignment, used for evaluatin of alignments *)
let remove_endspace map1 map2 =
  let f = ref 0 in
  while map1.(!f) = None || map2.(!f) = None do incr f; done;
  let l = ref ((Array.length map1)-1) in
  while map1.(!l) = None || map2.(!l) = None do decr l; done;
  let first,len = !f,(!l - !f +1) in
  (Array.sub map1 first len), (Array.sub map2 first len);;

let set_params ?(scores) ?(al = !altype)
    ?(go = !gap_open) ?(ge = !gap_ext) () =
  let sf = Util.q2list Scoring.scorefuncs in
  begin match scores with 
    None -> ()
  | Some(scores) -> 
      Scoring.reset ();
      List.iter Scoring.add_score scores
  end;
  let oal,ogo,oge = !altype, !gap_open, !gap_ext in
  altype := al; gap_open := go; gap_ext := ge;
  sf,oal,ogo,oge

let reset_params (scores,newal,ngo,nge) =
  Scoring.reset ();
  List.iter Scoring.add_score scores;
  altype := newal; gap_open := ngo; gap_ext := nge

open Ckarg
let options = [
  "-local","", "Use Smith-Waterman local alignment (default)",
    Unit(fun ()->altype:=Local);
  "-global","","Needleman-Wunsch global alignment",
    Unit(fun ()->altype:=Global);
  "-globalEF","","Needleman-Wunsch global alignment with endspace free",
    Unit(fun ()->altype:=GlobalEF);
  "-go","float","Gap opening cost",Set_float(gap_open);
  "-ge","float","Gap extension cost",Set_float(gap_ext);
  "-zscore","","Shuffle alignments, calculate z-score (default off)",Set(zscore);
  "-nshuffles","int","Number of times to shuffle for z-score",Set_int(nshuffles);
]

