(** Main entry point for counting columns, conservation *)

open Printf
open Align
open Sequence
open Util
open Scoring
open Matrices
open Align_io
open Query_align
open Colcount

let topk = ref 10				 (* Number of hits to consider *)
let exclude_queries = ref false			 (* Whether or not exclude queries *)
let thresh = ref 0.0 				 (* Threshold for converting indices *)
let wgt_scheme = ref `Equal			 (* Scheme to weight positions/hits for counting *)
let cnt_filt = ref never_cnt_filt		 (* How to score for local weights *)
let write_output = ref true		 	 (* Whether to write output *)
let idxname = Sequence.idxname			 (* Which index to use for labels *)
  
let local_width = ref 2

let query_keys = ref [||]			 (* Keys of queries *)

open Ckarg  
(** Options for program execution *)
let prog_opts = [
  "-topk","int","Max number of alignments for each query. May be less than original topk but not more.",
    Set_int(topk);
  "-queries","'key1 [key2 ...]'",
    "Select the queries to search for in the DB.  default = whole sequence database",
    String(fun s->query_keys:=Util.parse_strings s);
  "-cnt_filt","key","Filter some columns out of counting",SymbolArb(true,[
      "never","No filtering, always count",(fun () ->cnt_filt := never_cnt_filt);
      "query","Count only if percent gaps in query window is below tolerance",(fun ()->cnt_filt:= query_window_cnt_filt); 
      "db","Count only if percent gaps in db entry window is below tolerance",(fun ()->cnt_filt:= db_window_cnt_filt); 
      "both","Count only if percent gaps in both query/db window is below tolerance",(fun ()->cnt_filt:= both_window_cnt_filt); 
      "either","Count only if percent gaps in either query/db window is below tolerance",(fun ()->cnt_filt:= either_window_cnt_filt); 
    ]);
  "-filt_width","int","Width of window for window-based column filtering, must be >= 0",
    Set_int(Colcount.cnt_filt_width);
  "-filt_tol","float","Tolerance for gaps in window, lower means less tolerance, must be in [0,1]",
    Set_float(Colcount.cnt_filt_tol);
  "-wgt_scheme","key","Weighting scheme for topk matches.",
    SymbolArb(true,[
      "equal","Each topk hit is weighted equally",(fun ()-> wgt_scheme := `Equal);
      "global","Hits are weighted by how well they score against the query, \
         scores taken from the original alignment",(fun ()-> wgt_scheme:= `Global);
      "local","Weights depend on alignment score in a window around query/DB",
        (fun ()-> wgt_scheme:=`Local);
    ]);
  "-local_width","int","Width of window for local weight determination",
    Set_int(local_width);
  "-no_output","","Do not produce any output files (.thresh, .msa)",Clear(write_output);
  "-idxname","string","Which index to use for labels of residues",Set_string(idxname);
  "-dir","directory","Add the directory to search path for data files",String(Find.add_dir);
]

let options = prog_opts @ Seqdb.options @ Scoring.options @ Align.options

let usg = 
  sprintf "usage: %s [options]  -{filetype} seqfile -index indexfile [-{filetype} seqfile2 ...]"
    (Filename.basename Sys.executable_name)
let deffn s = failwith ("All arguments must have an option, "^s^" does not")

(* MAIN *)
let _ = begin
  parse options deffn usg;
  let topk, cnt_filt, write_output = !topk, !cnt_filt, !write_output  in
  let seqDB = Seqdb.seqDB in
  let nDB = Hashtbl.length seqDB in

  (* Verify the sequence DB *)
  ignore(Seqdb.verify_db seqDB);

  let query_keys = !query_keys in
  let nq = Array.length query_keys in
  printf "%d DB entries imported, %d queries selected\n" nDB nq;
  flush_all ();

  (* Add default scoring *)
  Scoring.default_score (seq_gsm_scores ~scale:(!Scoring.wgt) !Matrices.gmat);
  
  let maxlen = Seqdb.maxlen_seqDB seqDB in
  let scoremat = match !wgt_scheme with		 (* Score matrix, used with local wgt scheme *)
      `Equal | `Global -> [||]
    | `Local -> Array.make_matrix maxlen maxlen 0.0
  in
  let weights = Array.make_matrix maxlen topk (1.0 /. (float topk)) in

  let all_indexes = ref [] in
  let all_abst = ref [] in
  let all_ignt = ref [] in
  let all_gapt = ref [] in

  (* Loop over all queries *)
  for i=0 to nq-1 do
    let qkey = query_keys.(i) in
    let fname = Find.file (qkey ^ ".topk") in
    let f = open_in fname in
    let best : (float * topk_hit) array = Pervasives.input_value f in
    close_in f;

    (* How many hits to really check against *)
    let nbest = Array.length best in
    let truek = 
      if nbest < topk then 
	eprintf "Warning: %s has only %d top-k entries, truncating from %d\n%!" qkey nbest topk;
      min topk nbest
    in
    let best = Array.sub best 0 truek in 
    let qseq = Hashtbl.find seqDB qkey in
    let qindex = Sequence.get_index ~key:qkey qseq !idxname in

    (* Compute MSA and weights *)
    let msa = colcount_msa cnt_filt qseq !idxname best seqDB in
    let sasa_msa = sasa_msa cnt_filt qseq !idxname best seqDB in

    (* Fill in the weights *)
    begin match !wgt_scheme with
	`Equal -> () 				 (* Already all 1's *)
      | `Global -> global_msa_weights weights qseq best
      | `Local -> 
	  local_msa_weights weights !local_width scoremat qseq best seqDB
    end;

    let abst, ignt, gapt = colcount_thresh weights truek msa in
    all_indexes := qindex :: !all_indexes;
    all_abst := abst :: !all_abst;
    all_ignt := ignt :: !all_ignt;
    all_gapt := gapt :: !all_gapt;

    (* Output the thresholds to a file *)
    if write_output then begin
      write_out_msa string_of_int qindex msa (qkey^".msa");
(*       let f = Array.map float_of_int qindex in *)
(*       write_out_msa string_of_float f sasa_msa (qkey^".smsa"); *)
      write_sasa_consistency sasa_msa (qkey^".cons");
      write_gksasa_consistency sasa_msa (qkey^".gkcons");
      
      let f = open_out (qkey ^ ".thresh") in
      for i=0 to qseq.seqlen-1 do
	fprintf f "%d %f %f %f\n" qindex.(i) abst.(i) ignt.(i) gapt.(i);
      done;
      close_out f;
      let f = open_out (qkey ^ ".hts") in (* Output homology transfer scores *)
      Array.iter (fun x->fprintf f "%f\n" x) abst;
      close_out f;
      Util.write_matrix ~fname:(qkey ^ ".wgt") ~m:(qseq.seqlen) ~n:(truek) "%f " weights;
    end;
  done;

  (* Utility function to compute cumulative ROC using List.fold2*)
  let calcfrom2 cumroc va cl = 
    let vacl = Util.pair_arrays va cl in
    let roc, _,_,_,_,_ = Roc.calc_roc vacl in
    roc +. cumroc
  in
  let nq = float nq in

  (* Average, per protein ROCs *)
  let abs_roc = (List.fold_left2 calcfrom2 0.0 !all_abst !all_indexes) /. nq in
  
let ign_roc = (List.fold_left2 calcfrom2 0.0 !all_ignt !all_indexes) /. nq in
  let gap_roc = (List.fold_left2 calcfrom2 0.0 !all_gapt !all_indexes) /. nq in
  
  let all_indexes = Array.concat !all_indexes in
  let all_abst = Array.concat !all_abst in
  let all_ignt = Array.concat !all_ignt in
  let all_gapt = Array.concat !all_gapt in
  
  (* Overall ROCs and uniqueness *)
  let abs_rocall, _,_,_,_,_ = Roc.calc_roc (Util.pair_arrays all_abst all_indexes) in
  let ign_rocall, _,_,_,_,_ = Roc.calc_roc (Util.pair_arrays all_ignt all_indexes) in
  let gap_rocall, _,_,_,_,_ = Roc.calc_roc (Util.pair_arrays all_gapt all_indexes) in

  printf "ROCs\n";
  printf "%10s %10s %10s\n" "Style" "Per Prot" "Overall";
  printf "%10s %10.6f %10.6f\n" "Absolute"   abs_roc abs_rocall;
  printf "%10s %10.6f %10.6f\n" "No Ignore"  ign_roc ign_rocall;
  printf "%10s %10.6f %10.6f\n" "No Gap/Ign" gap_roc gap_rocall;
end

