(** Main entry point for aligning queries agains a database *)

open Printf
open Align
open Sequence
open Util
open Scoring
open Matrices
open Align_io
open Query_align

let topk = ref 10
let exclude_queries = ref false
let seq_filt = ref 1.0

let query_keys = ref [||]
let exclude_keys = ref [||]

open Ckarg
(** Options for program execution *)
let prog_opts = [
  "-topk","int","Max number of alignments for each query",Set_int(topk);
  "-queries","'key1 [key2 ...]'","Select the queries to search for in the DB",
    String(fun s->query_keys:=Util.parse_strings s);
  "-exclude","'key1 [key2 ...]'","Keys to be excluded from DB search",
    String(fun s->exclude_keys:=Util.parse_strings s);
  "-exclude_queries","","Exclude *all* queries from search",
    Set(exclude_queries);
  "-seq_filt","float",
    "If query and DB entry are above the identity threshold, exclude from search. \
     This simulates a DB which has sequence independence from the queries.",
    Set_float(seq_filt);
  "-dir","directory","Add the directory to search path for data files",String(Find.add_dir);
]

let options = prog_opts @ Scoring.options @ Align.options @ Seqdb.options

let usg = 
  sprintf "usage: %s [options] -queries 'q1 [q2 ..]' -{filetype} seqfile -index indexfile [-{filetype} seqfile2 ...]"
    (Filename.basename Sys.executable_name)
let deffn s = failwith ("All arguments must have an option, "^s^" does not")

(* MAIN *)
let _ = begin
  let startuptime = Timing.interval () in
  parse options deffn usg;
  let seqDB = Seqdb.seqDB in
  let query_keys = !query_keys in
  let nq,nDB = Array.length query_keys, Hashtbl.length seqDB in
  Hashtbl.iter verify_seq seqDB;

  Scoring.default_score (seq_gsm_scores ~scale:!wgt !gmat);

  (* Allocate space for alignments/results *)
  let maxlen = Seqdb.maxlen_seqDB seqDB in
  let alspace = Align.allocate_max maxlen in
  let topkpq = Query_align.make_topk_pq !topk in

  (* Hash to check if an DB entry should be ignored  *)
  let nexcl = Array.length !exclude_keys in
  let exclhash =			
    let h = Hashtbl.create nexcl in
    Array.iter (fun k->Hashtbl.add h k ()) !exclude_keys;
    h
  in

  (* Report startup results *)
  printf "%d DB entries imported, %d queries selected, %d entries excluded\n" 
    nDB nq (Hashtbl.length exclhash) ;
  flush_all ();

  let total_aligntime = ref 0.0 in
  let loadtime = Timing.interval () in

  (* Loop over all queries *)
  for i=0 to nq-1 do
    Heap.Minheap.empty topkpq;

    (* Get the best alignments *)
    let qkey = query_keys.(i) in
    printf "%4d: %s " i query_keys.(i); flush stdout;
    let best = 
      Query_align.get_top_aligns !seq_filt seqDB 
	topkpq alspace exclhash qkey
    in
    (* Output binary top best *)
    let out = open_out (qkey^".topk") in
    Pervasives.output_value out best;
    close_out out;
    (* Output readable list of top best *)
    let out = open_out (qkey^".list") in
    for j=0 to (Array.length best)-1 do
      let score,hit = best.(j) in
      fprintf out "%s %6d %f\n" hit.dbkey hit.alen score;
    done;
    close_out out;
    let score,hit = best.(0) in
    let curaligntime = Timing.interval () in
    printf "%s %.2f %.2f\n" hit.dbkey score curaligntime;
    total_aligntime := !total_aligntime +. curaligntime;
  done;
  printf "Timing\n%10s: %8.2f\n%10s: %8.2f\n%10s: %8.2f\n"
    "Startup" startuptime "DB Load" loadtime "Alignment" !total_aligntime;
end
