(** Main entry point for alignment tool program *)
open Printf

open Align
open Sequence
open Cmd_util
open Util
open Scoring
open Matrices
open Align_io

(** What type of operation to perform *)
type runtype = 
    AlignSeqs 				(** Align two sequences *)
  | Correspond				(** Hack to get prediction correspondences for CASP *)
  | Evaluate of string*string		(** Evaluate a precomputed alignment *)
  | Compare of string*string*string*string (** Compare two aligments *)

let runtype = ref AlignSeqs		(* Aligning, evaluating, comparing... *)
let seq1files = Queue.create ()		(* Files for seq1 *)
let seq2files = Queue.create ()		(* Files for seq2 *)
let endspace_free = ref false		(* Evaluating for endspace free? *)
let seq1 = ref (new_seq ())		(* Build seq1 *)
let seq2 = ref (new_seq ())		(* Build seq2 *)
let thresh = ref 0.0 			(* Threshold for continuous input *)
let output_scoremat = ref false		(* Whether to store the scoring matrix *)
let outformat = ref Standard		(* Format of output *)
let outseq = ref AAsequence		(* Type of sequence to output in Standard *)
let trans_table = ref "01" 		(* Default Translation table for indexes *)
let corrout = ref false			(* Whether to output alignment of sequence corrections *)
let idxname = Sequence.idxname		(* Alias for index name in Sequence module *)

let output_scoremat_fname = "scores"

let deffn s = 
  eprintf "Error: argument %s not preceded by an option, exiting\n" s;
  exit 1;;

let usg = sprintf "usage: %s [options] -{filetype} seqfile1 seqfile2"
  (Filename.basename Sys.executable_name)

(** Add a scoring scheme to list to that is to be used *)
let addscore func = Queue.add func scorefuncs;;

(** Read two sequences, update the global seq *)
let read2seqs t f1 f2 =
  Queue.add f1 seq1files;
  Queue.add f2 seq2files;
  let thresh = !thresh in
  seq1 := read_seq t ~thresh !seq1 f1;
  seq2 := read_seq t ~thresh !seq2 f2;;

(** Read a sequence of floating point numbers and convert to and index
    for the first sequence, read an index for the second sequence *)
let read_thresh2true f1 f2 =
  Queue.add f1 seq1files;
  Queue.add f2 seq2files;
  let thresh = !thresh in
  seq1 := read_seq Threshold ~thresh !seq1 f1;
  seq2 := read_seq Index ~thresh !seq2 f2;;

let setup_compare_align lst = 
  match lst with [atype1;fname1;atype2;fname2] ->
    runtype := Compare(atype1,fname1,atype2,fname2)
  | _ -> failwith "That was unexpected, wasn't it?";;

open Ckarg
let read_opts = [
  "-profile","file1 file2","Use PSI-blast profiles",String2(read2seqs Profile);
  "-fasta","file1 file2","FASTA files",String2(read2seqs FASTA);
  "-gkprofile","file1 file2","Profiles from PSI-Blast according the George's Modification",
    String2(read2seqs GKprofile);
  "-ss","file1 file2","Read secondary structure from files in FASTA format, H:helix, E:strand, C:coil",
    String2(read2seqs SSfasta);
  "-yasspp","file1 file2","YASSPP output file",String2(read2seqs YASSPP);
  "-yasspptrue","file1 file2","YASSPP output file, use predicted as true secondary structure",
    String2(read2seqs YASSPPtrue);
  "-index","file1 file2","Read a a vector of integers describing each AA of the sequence giving it current label",
    String2(read2seqs Index);
  "-idxname","string","Set the name of the next index",
    Set_string(idxname);
  "-thresh","file1 file2","Read floating point values, below t becomes 0, above becomes 1 in an index with current label. Useful for prediction inputs.",
    String2(read2seqs Threshold);
  "-thresh2true","pred true","First file is a threshold, second is an index. Useful for comparing predicted to true values",
    String2(read_thresh2true);
  "-t","float","Set the weight of thresholding for -thresh and -thresh2true",Set_float(thresh);
  "-floats","file1 file2","Read floating point features of a the sequences",
    String2(read2seqs Floats);
  "-dir","directory","Add the directory to search path for data files",String(Find.add_dir);
]    

let set_opts = [
  "-correspond","","Get correspondence between sequences, attributes are put in floats and printed in column format",
    Unit(fun ()->runtype:= Correspond);
  "-eval","type file","Evaluate alignment in file, useful for file format conversion.",
     String2(fun a f->runtype:= Evaluate(a,f));
  "-compare","type1 file1 type2 file2","Compare two alignments",
     StringList(4,setup_compare_align);
  "-o","key","Output format",SymbolArb(true,[
      "standard","Standard format, includes score",(fun ()->outformat:=Standard);
      "pir","PIR sequence format",(fun ()->outformat:=PIR);
      "c7pred","c7pred output format",(fun ()->outformat:=C7pred);
     ]);
  "-outseq","key","Style of output format",SymbolArb(false,[
      "aa","Amino acid sequence",(fun ()->outseq:=AAsequence);
      "ss","Secondary structure sequence",(fun ()->outseq:=SSsequence);
      "sspred2true","Secondary structure sequence",(fun ()->outseq:=SSpred2true);
      "index","Last index provided, must set the translation table using -trans",(fun ()->outseq:=IndexSeq(!idxname));
    ]);
  "-trans","string","Set the translation table to be used for an index",Set_string(trans_table);
  "-endspace_free","","Only for evaluation alignments, treat endspace as free",Set(endspace_free);
  "-output_scoremat","","Leave the computed DP scoring matrix in file 'scores' for debuggin purposes",Set(output_scoremat);
  "-corrout","","When evaluating a sequence alignment, output the derived sequence correspondence in seq1.corr and seq2.corr",Set(corrout);
]

let options = 
  set_opts @ read_opts @ Scoring.options @ Align.options

let gscoremat = ref [||]

(* MAIN *)
let _ = begin
  parse options deffn usg;
  let seq1,seq2 = !seq1,!seq2 in
  (* Default scoring is GSM *)
  Scoring.default_score (seq_gsm_scores ~scale:(!Scoring.wgt) !Matrices.gmat);
  begin match !runtype with
  | AlignSeqs ->
      if seq1=empty_seq ||seq2=empty_seq then
	(eprintf "No sequence files for one sequence\n"; exit 1);
      let xid,yid = 
	Util.basename (Queue.top seq1files),Util.basename (Queue.top seq2files) in
      verify_seq xid seq1;
      verify_seq yid seq2;
      let score,allen,map1,map2,(_,_,_,_,scoremat) = 
	allocate_align seq1 seq2 in
      let seqtrans = Align_io.get_outseq_array seq1 seq2 !outseq in
      print_alignment !outformat ~ids:(xid,yid) seqtrans map1 map2 score allen;
      gscoremat := scoremat;

  | Correspond ->
      if seq1=empty_seq ||seq2=empty_seq then
	(eprintf "No sequence files for one sequence\n"; exit 1);
      let xid,yid =
	Util.basename (Queue.top seq1files),Util.basename (Queue.top seq2files) in
      verify_seq xid seq1;
      verify_seq yid seq2;
      let corr = Align_io.find_seq_correspondence ~fout:(Some(xid^"-"^yid^".cor")) seq1 seq2 in
      if seq1.floats<>[||] && seq2.floats<>[||] then begin
	for i=0 to seq1.seqlen-1 do
	  match corr.(i) with
	  | Some(yi) -> printf "%g %g\n" seq1.floats.(i) seq2.floats.(yi)
	  | _ -> ();
	done;
      end
      else eprintf "No tclass attribute for sequences\n"


  | Evaluate(atype,afile) -> 
      let seq1,seq2,map1,map2 =
	Align_io.read_and_sync_align2seq ~corrout:!corrout atype afile seq1 seq2 in
      let xid,yid = split_filename afile in
      verify_seq xid seq1;
      verify_seq yid seq2;
      let map1,map2 = 
	if !endspace_free 
	then remove_endspace map1 map2
	else map1,map2 in
      let allen = Array.length map1 in
      let score,scoremat = 
	eval_alignment seq1 seq2 map1 map2 allen in
      let seqtrans = Align_io.get_outseq_array seq1 seq2 !outseq in
      print_alignment !outformat ~ids:(xid,yid) seqtrans map1 map2 score allen;
      gscoremat := scoremat;

  | Compare(atype1,afile1,atype2,afile2) -> 
      if seq1=empty_seq || seq2=empty_seq then
	(eprintf "No sequence files for one sequence\n"; exit 1);
      let xid,yid = Util.basename afile1, Util.basename afile2 in
      verify_seq xid seq1;
      verify_seq yid seq2;
      let _,_,xmap1,ymap1 = 
	read_and_sync_align2seq ~corrout:!corrout atype1 afile1 seq1 seq2 in
      let _,_,xmap2,ymap2 = 
	read_and_sync_align2seq ~corrout:!corrout atype2 afile2 seq1 seq2 in
      let nseq1 = alignment2numeric_seq seq1.seqlen seq2.seqlen xmap1 ymap1 in
      let nseq2 = alignment2numeric_seq seq1.seqlen seq2.seqlen xmap2 ymap2 in
      let tmp1 = {(new_seq ()) with seqlen=Array.length nseq1; seq=nseq1} in
      let tmp2 = {(new_seq ()) with seqlen=Array.length nseq2; seq=nseq2} in
      let compplan = [seq_match_scores ~scale:1.0] in
      let _ = 
	Align.set_params ~scores:compplan ~al:Align.Global ~go:0.0 ~ge:0.0 () in
      let score,allen,map1,map2,(_,_,_,_,scoremat) = allocate_align tmp1 tmp2 in
      print_compare_alignment ~ids:(xid,yid) 
	seq1 seq2 map1 map2 score allen nseq1 nseq2 xmap1 ymap1 xmap2 ymap2;
      gscoremat := scoremat;
  end;      

  (* Output the scoring matrix if desired *)
  if !output_scoremat then begin
    Util.write_matrix ~fname:output_scoremat_fname "%8.3f " !gscoremat;
  end
end
