open Util
open Printf

type flag = string			(* -option string *)
type exmp = string 			(* Example string *)
type doc = string			(* Some sort of documentation *)
type usage_msg = string
type anon_fun = (string -> unit)
type key = string
type init = bool  
  
type spec =
  | Unit	of (unit -> unit)     
  | Bool	of (bool -> unit)
  | Set		of (bool ref)
  | Clear	of (bool ref)
  | String	of (string -> unit)
  | Set_string	of (string ref)
  | Int		of (int -> unit)
  | Set_int	of (int ref)
  | Float	of (float -> unit)
  | Set_float	of (float ref)
  | String2     of (string -> string -> unit)
  | StringList  of int * (string list -> unit)
  | Symbol      of init * (string -> unit) * ((key * doc) list)
  | SymbolArb   of init * ((key * doc * (unit -> unit)) list)
  | Header      
  | Help        of usage_msg * ((flag * doc * exmp * spec) list)

(*   | Rest of (string -> unit)   (\* Stop interpreting keywords and call the *)
(*                                   function with each remaining argument *\) *)

(*   | Tuple of spec list         (\* Take several arguments according to the *)
(*                                   spec list *\) *)

(** Utility to build a list of strings for [StringList] *)
let rec build_stringlist cur stop lst =
  if cur < stop then
    build_stringlist (cur+1) stop (Sys.argv.(cur) :: lst)
  else
    List.rev lst;;

(** Get association based on first in tuple, return second pair *)
let rec assoc3 x l =
  match l with
  | [] -> raise Not_found
  | (y1, y2, y3) :: _ when y1 = x -> y2,y3
  | _ :: t -> assoc3 x t
;;

(** Split a 3 way list *)
let rec split3 l =
  match l with
	[] -> [],[],[]
  | (a,b,c)::t -> 
	  let al,bl,cl = split3 t in
	  (a::al,b::bl,c::cl)

(** Indent levels for docs *)
let indents = [|
  "\n  ";
  "\n     ";
  "\n       ";
|]

(** Widths for docs *)
let widths = [| 77; 74; 72 |]


(* let indent1=;; *)
(* let width1 = 75; *)
(* (\** Indent level for lists of Symbol keys *\) *)
(* let indent2="\n     ";; *)
(* (\** Indent level for descriptions of Symbol keys *\) *)
(* let indent3="\n      ";; *)


(** Figure out a default string for a spec *)
let get_default spec =
  let defs = "default: " in
  let nodef = "" in
  match spec with
  | Unit _ | Bool _ | Set _ | Clear _ | String _ | Int _
  | Float _ | String2 _ | StringList _ | Header | Help _ -> nodef

  | Set_string r -> defs ^ !r
  | Set_int r -> defs ^ (string_of_int !r)
  | Set_float r -> defs ^ (string_of_float !r)

  | Symbol (init, _, kd_list) ->
      if init then
	let k, _ = List.hd kd_list in
	defs ^ k
      else nodef
  | SymbolArb (init, kdf_list) ->
      if init then
	let k, _, _ = List.hd kdf_list in
	defs ^ k
      else nodef
;;


(** Print the long format documentation *)
let print_docs usage options =
  printf "%s\nOptions are:\n" usage;
  let iter (flag, exm, doc, spec) =
    let doc = wordwrap ~sep:indents.(0) widths.(0) doc in
    let default = get_default spec in
    let rest = match spec with
	Symbol (_, _, kd_list) ->
	  let getdoc (k, d) = k ^ indents.(2) ^ (wordwrap ~sep:indents.(2) widths.(2) d) in
	  (String.concat indents.(1) (List.map getdoc kd_list)) ^ "\n"
      | SymbolArb (_, kdf_list) ->
	  let getdoc (k, d, _) = k ^ indents.(2) ^ (wordwrap ~sep:indents.(2) widths.(2) d) in
	  (String.concat indents.(1) (List.map getdoc kdf_list)) ^ "\n"
      | Unit _	| Bool _	| Set _		| Clear _| String _ | Int _
      | Float _	| String2 _	| StringList _	| Set_string _      | Set_int _	
      | Set_float _	| Header | Help _ -> ""
    in
    if spec = Header then
      printf "**** %s ****\n\n" flag
    else
      printf "%s %s   %s%s%s%s%s\n" flag exm default indents.(0) doc indents.(1) rest
  in
  List.iter iter options;
;;

	
(** Initialize options, only relevant for [Symbol] and [SymbolArb] *)
let init_options options =
  let iter (_,_,_, spec) =
    match spec with
      Symbol (true, f, kd_list) ->
	let k,_ = List.hd kd_list in
	f k;
    | SymbolArb (true, kdf_list) ->
	let _,_,f = List.hd kdf_list in
	f ();
    | Unit       _
    | Bool       _
    | Set        _
    | Clear      _
    | String     _
    | Set_string _
    | Int        _
    | Set_int    _
    | Float      _
    | Set_float  _
    | String2    _
    | StringList _
    | Symbol     _
    | SymbolArb  _
    | Header
    | Help       _ -> ()
  in
  List.iter iter options;
;;

(** Deal with a single [spec] type - carry out its actions *)
let process_spec i spec =
  match spec with
  | Header -> (); i+1
  | Help (usage, options) -> print_docs usage options; exit 0;
  | Unit f -> f (); i+1
  | Bool f -> f (bool_of_string Sys.argv.(i+1)); i+2
  | Set r -> r := true; i+1
  | Clear r -> r := false; i+1
  | String f -> f Sys.argv.(i+1); i+2
  | Set_string r -> r := Sys.argv.(i+1); i+2
  | Int f -> f (int_of_string Sys.argv.(i+1)); i+2
  | Set_int r -> r := (int_of_string Sys.argv.(i+1)); i+2
  | Float f -> f (float_of_string Sys.argv.(i+1)); i+2
  | Set_float r -> r := (float_of_string Sys.argv.(i+1)); i+2
  | String2 f -> f Sys.argv.(i+1) Sys.argv.(i+2); i+3
  | StringList (n, f) -> 
      let l = build_stringlist (i+1) (i+1+n) [] in
      f l;   i+n+1;
  | Symbol (_, f, kd_lst) ->
      let keys, _ = List.split kd_lst in
      let k = Sys.argv.(i+1) in
      if List.mem k keys then
	(f k; i+2)
      else begin
	eprintf "Error: option %s must does not recognize %s, must be one of {%s}\n"
	  Sys.argv.(i) k (String.concat "|" keys);
	exit 1;
      end
  | SymbolArb (_, kdf_list) ->
      let k = Sys.argv.(i+1) in
      try
	let _, f = assoc3 k kdf_list in
	f (); i+2;
      with Not_found ->
	let keys = List.map (fun (x,_,_) -> x) kdf_list in
	eprintf "Error: option %s does not recognize %s, must be one of {%s}\n"
	  Sys.argv.(i) k (String.concat "|" keys);
	exit 1;
;;

(** Run through the arguments parsing them according to the [options]
    list *)
let parse (options: (flag * exmp * doc * spec) list) deffn usage =
  let flags = Hashtbl.create (List.length options) in
  List.iter (fun (flag,_,_,spec) -> Hashtbl.add flags flag spec) options;
  Hashtbl.add flags "-help" (Help (usage, options));
  Hashtbl.add flags "--help" (Help (usage, options));
  init_options options;
  let argc = Array.length Sys.argv in
  let i = ref 1 in
  while !i < argc do 
    if Sys.argv.(!i).[0] <> '-' then 
      try
	deffn Sys.argv.(!i);
	incr i;
      with _ -> 
	eprintf "Error: Problem with default argument handling of %s\n" Sys.argv.(!i);
	exit 1
    else try
      let spec = Hashtbl.find flags Sys.argv.(!i) in
      let next = process_spec !i spec in
      i := next;
    with
      Not_found -> 
	eprintf "Error: option %s not recognized\n" Sys.argv.(!i);
	exit 1;
    | Invalid_argument s | Failure s ->
	eprintf "Error: Invalid argument, '%s' for option %s\n" s Sys.argv.(!i);
	exit 1;
  done;;

