(** Basic min heap and topk implementation *)

(** Types optimized for float keys *)
type key = float;;

exception Empty_heap

let left i = i*2 + 1;;
let right i = i*2 + 2;;
let parent i = (i-1)/2;;

(** Signature of a structure to be given to Gheap functor, specifying
    the [key] for key type and [comp] for the comparison function *)
module type Heap_constr = sig
  type key
  val comp: key -> key -> bool
end;;

(** Functor to create heaps with arbitrary key types *)
module Gheap = functor (Hc:Heap_constr) -> struct

  type key = Hc.key
  type 'a t = {contents:(key * 'a) array; mutable size:int; maxsize:int }

  (** Comparison function for this heap *)
  let comp = Hc.comp

  (** Create a priority queu *)
  let make_pq size ik iv : 'a t = 
    {contents=Array.make size (ik,iv); size=0; maxsize=size}

  (** Empty the heap (fast, but not destructive) *)
  let empty (heap:'a t) = heap.size<-0
 
  (** Retrieve min key value pair *)
  let heap_min h = if h.size>0 then h.contents.(0) else raise Empty_heap

  (** True if the key [k] will make it into the PQ *)
  let will_add h k =
    h.size < h.maxsize || let mk,_ = heap_min h in comp mk k

  (** Maintain the heap condition *)
  let rec heapify (heap:'a t) i =
    let h,n = heap.contents,heap.size in
    let ik,_ = h.(i) in
    let l,r = (left i,right i) in
    let smalli,smallk =
      if l < n then
	let lk,_ = h.(l) in
	if comp ik lk then i,ik else l,lk
      else i,ik in
    let smalli,_ =
      if r < n then
	let rk,_ = h.(r) in
	if comp smallk rk then smalli,smallk else r,rk
      else smalli,smallk in
    if smalli <> i then begin
      let tkv = h.(smalli) in
      h.(smalli) <- h.(i);
      h.(i) <- tkv;
      heapify heap smalli
    end else ()

  (** Bubble position i up if it is smaller/larger than its parents *)
  let rec bubble_up (heap:'a t) i =
    let h = heap.contents in
    let ik,iv = h.(i) in
    let p = parent i in
    let pk,pv = h.(p) in
    if comp ik pk then begin
      h.(i) <- pk,pv;
      h.(p) <- ik,iv;
      if p > 0 then bubble_up heap p else ();
    end else ()

  (** Keep track of top k finishers, opposite of comp *)
  let topk_insert (heap:'a t) xk xv =
    let h = heap.contents in
    let n = heap.maxsize in
    if heap.size < n then begin
      h.(heap.size) <- xk,xv;
      bubble_up heap heap.size;
      heap.size <- heap.size+1;
    end else
      let mk,_ = heap_min heap in
      if comp mk xk then begin
	h.(0) <- xk,xv;
	heapify heap 0;
      end else ()

  (** Compare key/value pairs, for exporting *)
  let kv_comp (k1,_) (k2,_) =
    let lt,gt = (comp k1 k2),(comp k2 k1) in
    if lt && gt then 0
    else if lt then -1 else 1

  (** Compare key/value pairs, for exporting *)
  let rev_kv_comp (k1,_) (k2,_) =
    let lt,gt = (comp k1 k2),(comp k2 k1) in
    if lt && gt then 0
    else if lt then 1 else -1



  (** Export heap contents as a sorted array of key/value pairs *)
  let export (heap:'a t) =
    let contents = Array.make heap.size (heap.contents.(0)) in
    Array.blit heap.contents 0 contents 0 heap.size;
    Array.sort kv_comp contents;
    contents

  (** Export heap as a pq by reversing its contents *)
  let pqexport (heap:'a t) =
    let contents = Array.make heap.size (heap.contents.(0)) in
    Array.blit heap.contents 0 contents 0 heap.size;
    Array.sort rev_kv_comp contents;
    contents


end;;

(** A min-heap: apply functor to 'less than/equal to' *)
module Minheap = Gheap (struct
  type key = float
  let comp : float -> float -> bool = ( <= )
end );;
 
(* module Maxheap = Gheap (struct *)
(*   type key = string *)
(*   let comp : string -> string -> bool = ( > ) *)
(* end) ;; *)



(* module Gheap = functor (Hc:Heap_constr) -> struct *)
(* (\* Maintain the heap condition *\) *)
(*   type key = Hc.key *)
(*   type 'a t = ((key * 'a) array) * int ref *)
(*   let comp = Hc.comp *)
(*   let rec heapify ((h,sz):'a t) i = *)
(*     let n = !sz in *)
(*     let ik,_ = h.(i) in *)
(*     let l,r = (left i,right i) in *)
(*     let smalli,smallk =    *)
(*       if l < n then  *)
(* 	let lk,_ = h.(l) in *)
(* 	if comp ik lk then i,ik else l,lk *)
(*       else i,ik in *)
(*     let smalli,_ = *)
(*       if r < n then  *)
(* 	let rk,_ = h.(r) in *)
(* 	if comp smallk rk then smalli,smallk else r,rk  *)
(*       else smalli,smallk in *)
(*     if smalli <> i then begin *)
(*       let tkv = h.(smalli) in *)
(*       h.(smalli) <- h.(i); *)
(*       h.(i) <- tkv; *)
(*       heapify (h,sz) smalli *)
(*     end else () *)

(* (\* Bubble position i up if it is smaller/larger than its parents *\) *)
(*   let rec bubble_up ((h,sz):'a t) i = *)
(*     let ik,iv = h.(i) in *)
(*     let p = parent i in *)
(*     let pk,pv = h.(p) in *)
(*     if comp ik pk then begin *)
(*       h.(i) <- pk,pv; *)
(*       h.(p) <- ik,iv; *)
(*       if p > 0 then bubble_up (h,sz) p else (); *)
(*     end else () *)

(* (\* Keep track of top k finishers, opposite of comp *\) *)
(*   let topk_insert ((h,sz):'a t) xk xv = *)
(*     let n = Array.length h in *)
(*     if !sz < n then begin *)
(*       h.(!sz) <- xk,xv; *)
(*       bubble_up (h,sz) !sz; *)
(*       incr sz; *)
(*     end else  *)
(*       let mk,_ = heap_min (h,sz) in *)
(*       if comp mk xk then begin *)
(* 	h.(0) <- xk,xv; *)
(* 	heapify (h,sz) 0; *)
(*       end else () *)

(*   let make_pq size ik iv : 'a t = (Array.make size (ik,iv)),(ref 0) *)
(* end;; *)
