(** Computes kmer counts based on sequence or protein blocks for one
    file.  Encoding is fixed so that if the kmer becomes too large,
    arithmetic overflow will occur.  Best to use this on 64-bit
    systems.  Gives output in svm-light like format. *)

open Sequence
open Printf

let max_kmer = 13
let k = ref 3
let alphasize = ref 0
let seq = ref [||]

(** Computes the kmer index for a numeric sequence. [ii] is the offset
    in the sequence and [k] is the kmer length *)
let kmer_to_index s off k =
  let x = ref 0 in
  let b = ref 1 in
  for i=0 to k-1 do
    x := !x + !b * (s.(off+i));
    b := !b * !alphasize;
  done;
  !x


open Ckarg
let options = [
  "-fasta","file","FASTA file",
    String(fun f-> seq:= (Sequence.read_seq FASTA empty_seq f).seq; alphasize := Sequence.nstdaa);
  "-pblocks","file","Protein blocks format (ints 0-15)",
    String(fun f-> seq:= Util.read_ints ~skip1:true f; alphasize := 16);
  "-kmer","int",(sprintf "Set kmer length (max = %d)" max_kmer),
    Set_int(k);
]

let deffn s = 
  eprintf "Error: argument %s not preceded by an option, exiting\n" s;
  exit 1

let usg = sprintf "usage: %s [options] -{filetype} seqfile"
  (Filename.basename Sys.executable_name)

let _ = begin
  parse options deffn usg;
  let seq = !seq in
  let k = !k in
  let h = Hashtbl.create 100 in
  let n = Array.length seq in
  for i=0 to n-k-1 do
    let idx = kmer_to_index seq i k in
    if not (Hashtbl.mem h idx) then
      Hashtbl.add h idx 0;
    let cur = Hashtbl.find h idx in
    Hashtbl.replace h idx (cur+1);
  done;
  let kv = Array.make (Hashtbl.length h) (0,0) in
  let i = ref 0 in
  let f k v = kv.(!i) <- (k,v); incr i in
  Hashtbl.iter f h;
  Array.sort compare kv;
  Array.iter (fun (k,v) -> printf "%d:%d " (k+1) v) kv;
  printf "\n";
end;;
