(** Implements functions related to database searching *)
open Heap
open Align
open Sequence

(** Type for top-k hits: db key, alignment length, two alignment arrays *)
type topk_hit = {
  dbkey:string;				(** Key of database hit *)
  alen:int;				(** Length of alignment *)
  qal:int option array;			(** Query alignment array *)
  dal:int option array;			(** Database alignment array *)
}
(* type topk_hit = string * int * int option array * int option array *)

(** An empty hit *)
let empty_topk_hit = {dbkey=""; alen=0; qal=[||]; dal=[||]}
(* let empty_topk_hit = ("",0,[||],[||]) *)

(** Make an empty top-k alignment priority queue *)
let make_topk_pq k = Heap.Minheap.make_pq k 0.0 empty_topk_hit;;

(** Find the top alignments for query with index [qidx] using
    preallocated space and priority queue [pq]. This variant forces
    the single query to be in database and does not allow alphanumeric
    keys to be used in exclusion, everying is based on indices. *)
let get_top_aligns seqfilt seqDB pq (scores,v,e,f,tr,xmap,ymap) exclhash xkey =
  let x = Hashtbl.find seqDB xkey in
  let alignadd ykey y =
    if xkey <> ykey && not (Hashtbl.mem exclhash ykey) then begin
      Scoring.setup_scoring_mat scores x y;
      let score,alen = align_seqs v e f tr xmap ymap scores x.seqlen y.seqlen in
      let exact_identity,_ = Align_io.seq_identity x.seq y.seq xmap ymap alen in
      let addthis = Minheap.will_add pq score && exact_identity <= seqfilt in
      match addthis with
	true ->
	  let xmc,ymc = Array.make alen None, Array.make alen None in
	  Array.blit xmap 0 xmc 0 alen;
	  Array.blit ymap 0 ymc 0 alen;
	  Minheap.topk_insert pq score {dbkey=ykey; alen=alen; qal=xmc; dal=ymc}; 
      | _ -> ();			(* Not saving anything *)
    end;
  in
  Hashtbl.iter alignadd seqDB; 
  let results = Minheap.pqexport pq in
  results;;

