(** Support functions for computations of ROC *)

(** Compute the fraction of pos/neg class at each successive
    cutoff. Class labels must be sorted according to values.  *)
let calc_tpr_fpr va cl =
  let n = Array.length cl in
  let npos = Array.fold_left (fun m i -> if i>0 then m+1 else m) 0 cl in
  let nneg = n-npos in
  let oonpos,oonneg = (1.0 /. (float npos)),(1.0 /. (float nneg)) in
  let fracpos = Array.make (n+1) 0.0 in
  let fracneg = Array.make (n+1) 0.0 in
  let change = Queue.create () in
(*   if cl.(0) > 0 then *)
(*     fracpos.(1) <- oonpos *)
(*   else *)
(*     fracneg.(1) <- oonneg; *)
(*   Queue.add 1 change; *)
  for i=1 to n do
    let j = i-1 in
    fracpos.(i) <- fracpos.(i-1); 
    fracneg.(i) <- fracneg.(i-1); 
    if cl.(j) > 0 then 
      fracpos.(i) <- fracpos.(i) +. oonpos
    else
      fracneg.(i) <- fracneg.(i) +. oonneg;
    if j=n-1 || va.(j) <> va.(j+1) then
      Queue.add i change;
  done;  
  let change = Util.q2array change in
  let fpr = Array.map (fun ii->fracneg.(ii)) change in
  let tpr = Array.map (fun ii->fracpos.(ii)) change in
  tpr, fpr, change;;

(** Compute the true positive rate from a confusion matrix (same as
    recall and sensitivity) *)
(* TPR = tp / (tp+fn) = tp / npos = Recall = Sensitivity *)
let tpr_comp conf = conf.(1).(1) /. (conf.(1).(0) +. conf.(1).(1))

(** Compute the false positive rate *)
(* FPR = fp / (fp+tn) = fp / nneg *)
let fpr_comp conf = conf.(0).(1) /. (conf.(0).(1) +. conf.(0).(0))

(** Compute the precision *)
(* Precision = tp / (tp+fp) *)
let prec_comp conf = conf.(1).(1) /. (conf.(1).(1) +. conf.(0).(1))

(** Compute accuracy *)
let accu_comp conf = 
  (conf.(0).(0)+.conf.(1).(1))
  /. (conf.(0).(0)+.conf.(1).(1)+.conf.(0).(1)+.conf.(1).(0));;

let comp_sym_funcs = [
  "tpr",tpr_comp;
  "fpr",fpr_comp;
  "precision",prec_comp;
  "accuracy",accu_comp;
  "recall",tpr_comp;
  "sensitivity",tpr_comp;
]
let comp_syms_string = 
  let ss,_ = List.split comp_sym_funcs in
  String.concat " " ss;;


let scan_calc xcomp ycomp va cl npos nneg =
  let n = Array.length cl in
  let conf = Array.make_matrix 2 2 0.0 in (* row:actual col:predict *)
  conf.(0).(0) <- float nneg;
  conf.(1).(0) <- float npos;
  let x, y = Array.make (n+1) 0.0, Array.make (n+1) 0.0 in
  let change = Queue.create () in
  x.(0) <- xcomp conf;
  y.(0) <- ycomp conf;
  for i=1 to n do
    let act = if cl.(i-1) <= 0 then 0 else 1 in
    conf.(act).(0) <- conf.(act).(0) -. 1.0;
    conf.(act).(1) <- conf.(act).(1) +. 1.0;
    x.(i) <- xcomp conf;
    y.(i) <- ycomp conf;
    if i=n || va.(i-1) <> va.(i) then
      Queue.add i change;
  done;
  let change = Util.q2array change in
  let x = Array.map (fun ii->x.(ii)) change in
  let y = Array.map (fun ii->y.(ii)) change in
  x, y, change;;

(** Compute the area under the curve defined by points [(x,y)]. Use
    the trapezoid rule for this *)
let area_under_curve x y =
  let n = Array.length x in
  let auc = ref 0.0 in
  for j=1 to n-1 do
    let i = j-1 in 
    let w = x.(j) -. x.(i) in
    if w > 0.0 then
      let h = y.(j) -. y.(i) in
      auc := !auc +. y.(i)*.w +. (0.5 *. h *. w)
    else ();
  done;
  !auc;;

(** High level curve function, only need to give *)
let calc_curve xcomp ycomp vacl =
  Array.sort Util.revcomp vacl;
  let cl = Array.map (fun (_,c) -> c) vacl in
  let va = Array.map (fun (v,_) -> v) vacl in
  let n = Array.length cl in
  let npos = Array.fold_left (fun m i -> if i>0 then m+1 else m) 0 cl in
  let nneg = n - npos in
  let x, y, change = scan_calc xcomp ycomp va cl npos nneg in
  let auc = area_under_curve x y in
  auc, x, y, cl, va, change;;

(** Compute an ROCN statistic. *)
let calc_rocn rocn xcomp ycomp vacl =
  Array.sort Util.revcomp vacl;
  let cl = Array.map (fun (_,c) -> c) vacl in
  let va = Array.map (fun (v,_) -> v) vacl in
  let n = Array.length cl in
  let npos = Array.fold_left (fun m i -> if i>0 then m+1 else m) 0 cl in
  let nneg = n - npos in
  let idx, cnt = ref 0, ref 0 in
  while !cnt < rocn && !idx < n do
    if cl.(!idx) <> 1 then
      incr cnt;
    incr idx;
  done;
  let van = Array.sub va 0 !idx in
  let cln = Array.sub cl 0 !idx in
  let x, y, change = scan_calc xcomp ycomp van cln npos nneg in
  let auc = area_under_curve x y in
  auc, x, y, cln, van, change;;

(** Only use the top N predictions for computing the curve. This is
    not an ROCN computation *)
let calc_curve_nbest n xcomp ycomp vacl =
  Array.sort Util.revcomp vacl;
  let vacl_nb = Array.sub vacl 0 n in
  calc_curve xcomp ycomp vacl_nb
;;

(** Specialize [calc_curve] to ROC *)
let calc_roc vacl =
  calc_curve fpr_comp tpr_comp vacl;;

(* (\** Untested, unremembered... *\) *)
(* let compute_accuracy tpr fpr npos = *)
(*   let n = Array.length tpr in *)
(*   let nneg = n-npos in *)
(*   let posconv = (float npos) /. (float n) in *)
(*   let negconv = (float nneg) /. (float n) in *)
(*   let acc = Array.make n 0.0 in *)
(*   for i=0 to n-1 do *)
(*     let a = tpr.(i) *. posconv +. (1.0 -. fpr.(i)) *. negconv in *)
(*     acc.(i) <- a; *)
(*   done; *)
(*   acc;; *)

(** Returns the percent unique elements of sorted array *)
let uniqueness va =
  let n = Array.length va in
  let nu = ref 0 in
  for i=1 to n-1 do
    if va.(i)<>va.(i-1) then incr nu;
  done;
  (float !nu)/.(float (n-1));;
