open Util
open Sequence
open Printf
open Matrices

(** Issue a warning message if sequence features are zero length for a
    matrix scoring routine *)
let zero_len_check m n s =
  if m=0 || n=0 then
    eprintf "Warning: %s, zero-length m = %d  n = %d\n" s m n;;

(** Make sure position is in bounds for single scores *)
let in_bounds i j m n s =
  let ok = i<m && j<n && i>=0 && j>=0 in
  if not ok then
    eprintf "Index out of bounds on %s, %d vs %d and %d vs %d" s i m j n;
  ok;;
  
(** Adds the PICASSO3 scores between sequences to the [scores] matrix.
    The PSFM is normalized which may result in nans for the score. *)
let picasso_scores ?(scale=1.0) x y scores =
  let ps1,ps2,pf1,pf2 = x.prof.pssm,y.prof.pssm,x.prof.psfm,y.prof.psfm in
  zero_len_check x.prof.len y.prof.len "picasso_scores";
  let sum_or_1 a = 
    let s = Util.array_sumf a in
    if s = 0.0 then 1.0
    else s
  in
  let pfnorm1 = Array.init x.seqlen (fun i-> 1.0 /. (sum_or_1 pf1.(i))) in
  let pfnorm2 = Array.init y.seqlen (fun i-> 1.0 /. (sum_or_1 pf2.(i))) in
  for i=0 to x.seqlen-1 do
    for j=0 to y.seqlen-1 do
      let s12 = (dotp ps1.(i) pf2.(j)) *. pfnorm2.(j) in
      let s21 = (dotp pf1.(i) ps2.(j)) *. pfnorm1.(i) in
      let s = s12 +. s21 in
      scores.(i).(j) <- scores.(i).(j) +. scale*.s;
    done;
  done;;

let shift_scores ?(scale=1.0) x y scores =
  for i=0 to x.seqlen-1 do
    for j=0 to y.seqlen-1 do
      scores.(i).(j) <- scores.(i).(j) +. scale;
    done;
  done;;

(** Single position picasso score *)
let picasso_single_score ?(scale=1.0) x y i j =
  let ps1,ps2,pf1,pf2 = x.prof.pssm,y.prof.pssm,x.prof.psfm,y.prof.psfm in
  assert(in_bounds i j x.prof.len y.prof.len "picasso_single_score");
  let pfnorm1 = Util.array_sumf pf1.(i) in
  let pfnorm2 = Util.array_sumf pf2.(j) in
  let s12 = (dotp ps1.(i) pf2.(j)) /. pfnorm2 in
  let s21 = (dotp pf1.(i) ps2.(j)) /. pfnorm1 in
  let s = s12 +. s21 in
  scale*.s;;

(** Adds the Normalized PICASSO3 scores between sequences to the
    [scores] matrix.  The PSFM is normalized which may result in nans
    for the score. *)
let picasson_scores ?(scale=1.0) x y scores =
  let ps1,ps2,pf1,pf2 = x.prof.pssm,y.prof.pssm,x.prof.psfm,y.prof.psfm in
  zero_len_check x.prof.len y.prof.len "picasso_scores";
  let picnrm1 = Array.make x.seqlen 0.0 in
  for i=0 to x.seqlen-1 do
    let scr = sqrt (2.0 *. (fabs (dotp ps1.(i) pf1.(i)))) in
    picnrm1.(i) <- if scr > 0.0 then scr else 1.0
  done;
  let picnrm2 = Array.make y.seqlen 0.0 in
  for i=0 to y.seqlen-1 do
    let scr = sqrt (2.0 *. (fabs (dotp ps2.(i) pf2.(i)))) in
    picnrm2.(i) <- if scr > 0.0 then scr else 1.0
  done;
  for i=0 to x.seqlen-1 do
    for j=0 to y.seqlen-1 do
      let s12 = (dotp ps1.(i) pf2.(j))  in
      let s21 = (dotp pf1.(i) ps2.(j))  in
      let s = (s12 +. s21) /. (picnrm1.(i) *. picnrm2.(j)) in
      scores.(i).(j) <- scores.(i).(j) +. scale*.s;
    done;
  done;;


(** Adds dot products of given features to [scores] *)
let dotp_scores ?(scale=1.0) x1 x2 scores =
  let m = Array.length x1 in
  let n = Array.length x2 in
  for i=0 to m-1 do
    for j=0 to n-1 do
      let s = dotp x1.(i) x2.(j) in
      scores.(i).(j) <- scores.(i).(j) +. scale*.s;
    done;
  done;;

(** Adds dot products of given features to [scores], specific to
    pssms *)
let pssm_dotp_scores ?(scale=1.0) x y scores =
  zero_len_check (Array.length x.prof.pssm) (Array.length y.prof.pssm) "pssm_dotp_scores";
  dotp_scores ~scale x.prof.pssm y.prof.pssm scores;;

(** Single position pssmdot score *)
let pssm_dotp_single_score ?(scale=1.0) x y i j =
  assert(in_bounds i j (Array.length x.prof.pssm) (Array.length y.prof.pssm) "pssm_dotp_single_score");
  scale *. (dotp x.prof.pssm.(i) y.prof.pssm.(j));;

(** Adds dot products of given features to [scores], specific to
    secondary structure *)
let ssdotp_scores ?(scale=1.0) x y scores =
  zero_len_check (Array.length x.ssprof) (Array.length y.ssprof) "ssdotp_scores";
  dotp_scores ~scale x.ssprof y.ssprof scores;;

(** Single position ssdotp score *)
let ssdotp_single_score ?(scale=1.0) x y i j =
  assert(in_bounds i j (Array.length x.ssprof) (Array.length y.ssprof) "ssdotp_single_score");
  scale *. (dotp x.ssprof.(i) y.ssprof.(j));;

(** Adds cos of given features to [scores] *)
let cos_scores ?(scale=1.0) x1 x2 scores =
  let m = Array.length x1 in
  let n = Array.length x2 in
  let x1nrms = Array.map Util.norm2 x1 in
  let x2nrms = Array.map Util.norm2 x2 in
  for i=0 to m-1 do
    for j=0 to n-1 do
      let s = (dotp x1.(i) x2.(j)) /. x1nrms.(i) /. x2nrms.(j) in
      scores.(i).(j) <- scores.(i).(j) +. scale*.s;
    done;
  done;;

(** Adds dot products of given features to [scores], specific to
    pssms *)
let pssm_cos_scores ?(scale=1.0) x y scores =
  zero_len_check (Array.length x.prof.pssm) (Array.length y.prof.pssm) "pssm_cos_scores";
  cos_scores ~scale x.prof.pssm y.prof.pssm scores;;

(** Single position pssmdot score *)
let pssm_cos_single_score ?(scale=1.0) x y i j =
  assert(in_bounds i j (Array.length x.prof.pssm) (Array.length y.prof.pssm) "pssm_cos_single_score");
  let px,py = x.prof.pssm.(i), y.prof.pssm.(j) in
  scale *. (dotp px py) /. (Util.norm2 px) /. (Util.norm2 py) ;;


(** Adds the scores from a global scoring matrix to [scores] *)
let gsm_scores ?(scale=1.0) gsm seq1 seq2 scores =
  let m = Array.length seq1 in
  let n = Array.length seq2 in
  for i=0 to m-1 do
    for j=0 to n-1 do
      let s = gsm.(seq1.(i)).(seq2.(j)) in
      scores.(i).(j) <- scores.(i).(j) +. scale*.s;
    done;
  done;;

(** Adds the scores from a global scoring matrix to [scores], specific
    to sequence *)
let seq_gsm_scores ?(scale=1.0) gsm x y scores =
  zero_len_check (Array.length x.seq) (Array.length y.seq) "seq_gsm_scores";
  gsm_scores gsm ~scale x.seq y.seq scores;;

(** Single position sequence GSM score *)
let seq_gsm_single_score ?(scale=1.0) gsm x y i j =
  assert(in_bounds i j (Array.length x.seq) (Array.length y.seq) "seq_gsm_single_score");
  scale *. gsm.(x.seq.(i)).(y.seq.(j));;

(** SS global score *)
let ss_gsm_scores gsm ?(scale=1.0) x y scores =
  zero_len_check (Array.length x.ss) (Array.length y.ss) "ss_gsm_scores";
  gsm_scores ~scale gsm x.ss y.ss scores;;

(** Single SS GSM score *)
let ss_gsm_single_score ?(scale=1.0) gsm x y i j =
  assert(in_bounds i j (Array.length x.ss) (Array.length y.ss) "ss_gsm_single_score");
  scale *. gsm.(x.ss.(i)).(y.ss.(j));;

(** Add global scoring accoring to an index to [scores] *)
let index_gsm_scores ?(scale=1.0) idxname gsm x y scores =
  let xindex,  yindex = 
    try (Sequence.get_index x idxname),(Sequence.get_index y idxname)
    with Failure(s) -> failwith ("index_gsm_scores:"^s)
  in
  zero_len_check (Array.length xindex) (Array.length yindex) "index_gsm_scores";
  gsm_scores ~scale gsm xindex yindex scores;;

(** Single GSM score *)
let index_gsm_single_score ?(scale=1.0) idxname gsm x y i j =
  let xindex,  yindex = 
    try (Sequence.get_index x idxname),(Sequence.get_index y idxname)
    with Failure(s) -> failwith ("index_gsm_scores:"^s)
  in
  assert(in_bounds i j (Array.length xindex) (Array.length yindex) "ss_gsm_single_score");
  scale *. gsm.(xindex.(i)).(yindex.(j));;

(** Adds matching scores to [scores] *)
let match_scores ?(scale=1.0) seq1 seq2 scores =
  let m = Array.length seq1 in
  let n = Array.length seq2 in
  for i=0 to m-1 do
    for j=0 to n-1 do
      let s = if seq1.(i)=seq2.(j) then 1.0 else 0.0 in
      scores.(i).(j) <- scores.(i).(j) +. scale*.s;
    done;
  done;;

(** Adds matching scores to [scores], specific to sequence *)
let seq_match_scores ?(scale=1.0) x y scores =
  zero_len_check (Array.length x.seq) (Array.length y.seq) "seq_match_scores";
  match_scores ~scale x.seq y.seq scores;;

(** Single position match score *)
let match_single_score ?(scale=1.0) x y i j =
  assert(in_bounds i j (Array.length x.seq) (Array.length y.seq) "seq_gsm_single_score");
  if x.seq.(i) = y.seq.(j) then scale else 0.0;;

(** Add secondary structure profile dot produce against constant
    (true) profile to [scores] *)
let ss_dotp_1const_scores ?(scale=1.0) defss x y scores =
  zero_len_check (Array.length x.ssprof) (Array.length y.ss) "ss_dotp_1const_scores";
  let ss1,ss2idx = x.ssprof,y.ss in
  let m = Array.length ss1 in
  let n = Array.length ss2idx in
  for i=0 to m-1 do
    for j=0 to n-1 do
      let s = dotp ss1.(i) defss.(ss2idx.(j)) in
      scores.(i).(j) <- scores.(i).(j) +. scale*.s;
    done;
  done;;

(** Single position ssprofile to constant ss score score *)
let ss_dotp_1const_single_score ?(scale=1.0) defss x y i j =
  assert(in_bounds i j (Array.length x.ssprof) (Array.length y.ss) "seq_gsm_single_score");
  scale *. (dotp x.ssprof.(i) defss.(y.ss.(j)));;

(** Add score of predicted residue class in [x] against true residue
    class in [y] according to scoring matrix [mat] *)
let pred_true_inner_prod ?(scale=1.0) mat x y scores =
  zero_len_check (Array.length x.pclass) (Array.length y.tclass) "ss_dotp_1const_scores";
  let xpred = x.pclass in
  let ytrue = y.tclass in
  for i=0 to x.seqlen-1 do
    for j=0 to y.seqlen-1 do
      let s = Util.inner_product mat xpred.(i) ytrue.(j) in
      scores.(i).(j) <- scores.(i).(j) +. scale*.s;
    done;
  done;;

(** Shift scoring matrix by a constant, scale is the shift *)
let shift_scores ?(scale=0.0) x y scores =
  for i=0 to x.seqlen-1 do
    for j=0 to y.seqlen-1 do
      scores.(i).(j) <- scores.(i).(j) +. scale;
    done;
  done;;

(** Keeps the scoring mechanism *)
let scorefuncs = Queue.create ()	

(** Apply scoring mechanisms to a scoring matrix *)
let setup_scoring_mat ?(reset=true) scores (x:sequence) (y:sequence) =
  if reset then 
    fill_matrix scores 0.0; (* Reset to zeros *)
  Queue.iter (fun f->f x y scores) scorefuncs;;

(** Add a scoring scheme to list to that is to be used *)
let add_score func = Queue.add func scorefuncs;;

(** Set the default scoring mechanism: if the scoring queue is empty,
    add the given function *)
let default_score func =   
  if Queue.is_empty scorefuncs then 
    Queue.add func scorefuncs

(** Empty the scoring queue *)
let reset () = Queue.clear scorefuncs

let single_score func = reset (); add_score func

(** Weights on the scoring function *)
let wgt = ref 1.0
(* (\** Current global sequence scoring matrix *\) *)
(* let gmat = ref Matrices.blosum62	 *)
(** Current new scoring matrix *)
let newmat = ref [||]			
(** Current index for sequences *)
let idxname = Sequence.idxname

open Ckarg

(** Command options for scoring.  Include this in the command options
    to allow scoring functions to be set from a program. *)
let options = [
  "-s","key","Add a score type",SymbolArb(false,[
      "gsm","Current global scoring matrix scores, available GSMs: BLOSUM62",
        (fun ()-> add_score (seq_gsm_scores ~scale:!wgt !Matrices.gmat));
      "picasso","Add on profile to profile scoring",
        (fun ()-> add_score (picasso_scores ~scale:!wgt));
      "picasson","Add on normalized picasso score",
        (fun ()-> add_score (picasson_scores ~scale:!wgt));
      "pssmdot","Dot product of pssms",
        (fun ()-> add_score (pssm_dotp_scores ~scale:!wgt));
      "pssmcos","Cosine between pssms",
        (fun ()-> add_score (pssm_cos_scores ~scale:!wgt));
      "ssmat","Use a global secondary structure scoring matrix",
        (fun ()->add_score (ss_gsm_scores ~scale:!wgt Matrices.ssmat));
      "ssprofile","Use dot products of secondary structure profiles",
        (fun ()->add_score (ssdotp_scores ~scale:!wgt));
      "ssprof2true","Use dot products, but for the second sequence, use the averag secondary structure profile for its true SS",
        (fun ()->add_score (ss_dotp_1const_scores ~scale:!wgt Matrices.defss_prof));
      "shift","Shift scoring matrix by the weight assigned (-wgt 2 -s shift gives a shift of 2)",
        (fun ()->add_score (shift_scores ~scale:!wgt));
      "match","Sequence matching characters score 1, mismatch is 0",
        (fun ()-> add_score (seq_match_scores ~scale:!wgt));
      "index","Use the currently set up index and scoring matrix",
        (fun ()-> add_score (index_gsm_scores ~scale:!wgt !idxname !newmat));
    ]);
  "-newmat","file","Read a new global scoring mat for use with the next index",
    String(fun f->newmat:=read_matrix f);
  "-wgt","float","Weight on next scoring scheme",Set_float(wgt);
]
