(** This file contains a few utilities for working with sequence
    profiles. *)

open Util

let stdaaorder = "ARNDCQEGHILKMFPSTWYV" (** Single letter to integer conversion *)
let nstdaa = String.length stdaaorder	(** Number of standard AAs*)

let totaaorder = "ARNDCQEGHILKMFPSTWYVBZX*" (** Single letter to integer conversion *)
let ntotaa = String.length totaaorder	(** Total number of allowed AAs *)

(** Create a translation table for characters in [s] *)
let make_trans_table s = 
  let a = Array.make 256 (-1) in
  let n = String.length s in
  for i=0 to n-1 do
    a.(int_of_char s.[i]) <- i;
  done;
  a

(** Array of integers for the numeric codes of amino acids, TOTAL order used *)
let aacodes1 = make_trans_table totaaorder
(** Convert a single letter AA to a numeric code, standard only*)
let c2i c = aacodes1.(int_of_char c)
(** Convert a numeric index to the corresponding AA letter *)
let i2c i = totaaorder.[i]

(** Order of secondary structure elements *)
let ssorder = "CEH"
(** Number of secondary structure elements recognized *)
let nsstruct = String.length ssorder
(** Translation table for secondary structure *)
let sscodes = make_trans_table ssorder
(** Translate secondary structure to numeric indices *)
let ss2i c = sscodes.(int_of_char c)
(** Translate from index to ss character *)
let i2ss i = ssorder.[i]

(** Read from a FASTA file, potentially get multiple sequences *)
let read_fasta fname =
  let f = open_in fname in
  let seq_ids = Queue.create () in
  let seqs = Queue.create () in
  let s = Buffer.create 300 in
  let id = ref (input_line f) in
  begin try while true do
    let l = input_line f in
    if l.[0] = '>' then begin
      Queue.add !id seq_ids;
      Queue.add (Buffer.contents s) seqs;
      id := l;
      Buffer.clear s
    end
    else Buffer.add_string s l
  done
  with End_of_file | Invalid_argument(_) -> (* Accepting invalid_argument may make cause damage later *)
    Queue.add !id seq_ids;
    Queue.add (Buffer.contents s) seqs;
  end;
  close_in f;
  (q2array seqs),(q2array seq_ids)
  
(** Convert an amino sequence to numeric indices *)
let aa2idx s =
  Array.init (String.length s) (fun i -> c2i s.[i])
(** Convert a SS sequence to numeric indices *)
let ss2idx s =
  Array.init (String.length s) (fun i -> ss2i s.[i])

(** Convert an index array to a character string *)
let idx2aa a =
  let n = Array.length a in
  let s = String.create n in
  for i=0 to n-1 do s.[i] <- i2c a.(i); done;
  s;;

(** Type of a profile to hold matrices and vectors *)
type profile = {
  len:int;				(** Length of the profile *)
  sequence:int array;			(** Numeric sequence from the profile *)
  seqaa:string;				(** Sequence from the profile *)
  pssm:float array array;		(** Position specific scoring matrix *)
  psfm:float array array;		(** Position specific frequency matrix *)
  ipp:float array;			(** Information per position *)
  relwgt:float array;			(** Relative weight *)
}

let prof_size = 20			(** Number of columns in a PSSM/PSFM *)

(** An empty profile *)
let empty_profile : profile = {
  len=0;		
  sequence=[||];
  seqaa="";
  pssm=[||];
  psfm=[||];
  ipp=[||];
  relwgt=[||];
}


(** Take a slice of a profile *)
let profile_slice p start len =
  {len = len;
   sequence = Array.sub p.sequence start len;
   seqaa = String.sub p.seqaa start len;
   pssm = Array.sub p.pssm start len;
   psfm = Array.sub p.psfm start len;
   ipp = Array.sub p.ipp start len;
   relwgt = Array.sub p.relwgt start len;
  };;

(** Take a slice of the profile starting with [start] and ending with [stop] *)
let profile_slice_idx p start stop =
  let len = stop-start+1 in
  profile_slice p start len;;

(** Type to accumulate sequence information *)
type sequence = {
  seqlen:int;
  seq:int array;
  prof:profile;
  ss:int array;
  sspred:int array;
  ssprof:float array array;
  sasa:float array;
  indexes:(string , (int array)) Hashtbl.t;
  tclass:float array array;
  pclass:float array array;
  floats:float array;			(** Array of float features for random tasks *)
(*   matrices:(string * float array array) list; *)
};;

(** Make a new sequence *)
let new_seq () = {
  seqlen=0; 
  seq=[||]; 
  prof=empty_profile; 
  ss=[||]; 
  sspred=[||];
  ssprof=[||]; 
  sasa=[||];
  indexes=Hashtbl.create 5;
  tclass=[||];
  pclass=[||];
  floats=[||];
 };;

(** Default empty sequence *)
let empty_seq = new_seq ()

open Scanf

(** Read a single profile from file. If [gkform] is true, then the
    profile is George's modified PSI-BLAST output *)
let read_profile ?(gkform=false) fname =
  let nlines = Util.count_lines fname in
  let n = if gkform then nlines-1 else nlines-9 in
  let aaseq = String.create n in
  let pssm = Array.make_matrix n prof_size 0.0 in
  let psfm = Array.make_matrix n prof_size 0.0 in
  let ipp = Array.make n 0.0 in
  let relwgt = Array.make n 0.0 in

  let f = Scanning.from_file fname in
  bscanf f "%s@\n" ignore;
  if not gkform then
    bscanf f "%_s@\n%s@\n" ignore;
  for i=0 to n-1 do
    bscanf f " %_d %c" (fun c-> aaseq.[i]<-c);
    for j=0 to prof_size-1 do		(* PSSM *)
      pssm.(i).(j) <- bscanf f " %f " ident;
    done;
    for j=0 to prof_size-1 do		(* PSFM *)
      psfm.(i).(j) <- bscanf f " %f " ident;
    done;
    ipp.(i)    <- Util.rfloat f;
    relwgt.(i) <- Util.rfloat f;

  done;
  (* Scan to end of file, only way to close file handles with scanbufs *)
  while not (Scanning.end_of_input f) do
    bscanf f "%s@\n" ignore;
  done;
  {len=n; sequence=aa2idx aaseq; seqaa=aaseq;
   pssm=pssm; psfm=psfm; ipp=ipp; relwgt=relwgt}

open Printf

(** Write a PSI-BLAST style profile *)
let write_profile ?(gkform=false) fname prof =
  let f = open_out fname in
  let write form v = 
    for i=0 to 19 do
      fprintf f form (truncate v.(i));
    done;
    fprintf f " "; 
  in
  let write3 = write "%3d" in
  let write4 = write "%4d" in
  if not gkform then
    fprintf f "\nLast position-specific scoring matrix computed, weighted observed percentages rounded down, information per position, and relative weight of gapless real matches to pseudocounts\n";
  fprintf f "           A  R  N  D  C  Q  E  G  H  I  L  K  M  F  P  S  T  W  Y  V   A   R   N   D   C   Q   E   G   H   I   L   K   M   F   P   S   T   W   Y   V\n";
  for i=0 to prof.len-1 do
    fprintf f "%5d %c  " (i+1) prof.seqaa.[i];
    write3 prof.pssm.(i);
    write4 prof.psfm.(i);
    fprintf f "%4.2f %4.2f\n" prof.ipp.(i) prof.relwgt.(i);
  done;
  if not gkform then
    fprintf f "\n\
                      K         Lambda\n\
Standard Ungapped    0.0000     0.0000\n\
Standard Gapped      0.0000     0.0000\n\
PSI Ungapped         0.0000     0.0000\n\
PSI Gapped           0.0000     0.0000\n\
";;



(** Read a YASSP file and return the AA and SS sequence and the
	secondary structure profile *)
let read_yasspp fname =
  let n = count_lines fname in
  let f = Scanning.from_file fname in
  let aaseq = String.create n in
  let ssseq = String.create n in
  let ssprof= Array.make_matrix n nsstruct 0.0 in
  for i=0 to n-1 do
    aaseq.[i] <- bscanf f "%_d %c " ident;
    ssseq.[i] <- bscanf f "%_c %c " ident;
    for j=0 to nsstruct-1 do
      ssprof.(i).(j) <- bscanf f " %f " ident;
    done;
  done;
  aaseq,ssseq,ssprof

(** Get an index, raise [Not_found] if the index is not available *)
let get_index ?(key="Sequence") seq idxname =
  try
    Hashtbl.find seq.indexes idxname
  with Not_found ->
    failwith (sprintf "%s does not have index %s for labels" key idxname)

open Printf;;

type seqfile =
    FASTA
  | Profile
  | GKprofile				(** George's modified PSI-blast *)
  | YASSPP				(** YASSP output, fill in predss *)
  | YASSPPtrue				(** YASSP output, fill in true ss *)
  | SSfasta				(** CEH fasta, fill in true ss *)
  | SASA				(** Floats, solvent accesible surface area *)
  | Index				(** Integer data, space separated *)
  | Index1				(** Index with one-line header *)
  | Threshold				(** Floating data, space-separated *)
  | TClass				(** Classes of residue based on an index *)
  | P2Class				(** Predicted class, max 2, based on threshold*)
  | Floats				(** Float features *)
;;

(** Some default types based on extensions  *)
let ext_types = [
  ".fa", FASTA;
  ".profile", Profile;
  ".profile.gk", GKprofile;
  ".yasspp", YASSPP;
  ".ss", SSfasta;
  ".sasa", SASA;
  ".label", Index;
  ".dat",Floats;
]

let idxname = ref "index"

(** Read a sequence file and return a list of features read about the
    sequence. *)
let read_seq filet ?(thresh=0.0) seq fname =
  let fname = Util.Find.file fname in
  try begin match filet with
    FASTA -> 
      let aa,_  = read_fasta fname in
      {seq with seq=(aa2idx aa.(0)); seqlen=String.length aa.(0)}
  | Profile | GKprofile -> 
      let gkform = filet = GKprofile in
      let p = read_profile ~gkform fname in
      {seq with seq=(p.sequence); prof=(p); seqlen=p.len}
  | YASSPP ->
      let aa,ss,ssp = read_yasspp fname in
      {seq with 
	 seq=(aa2idx aa); sspred=(ss2idx ss); ssprof=(ssp); seqlen=String.length aa}
  | YASSPPtrue ->			(* Use YASSPP as true SS also  *)
      let aa,ss,ssp = read_yasspp fname in
      let ss = ss2idx ss in
      {seq with
	 seq=(aa2idx aa); ss=ss; sspred=ss; ssprof=(ssp); seqlen=String.length aa}
  | SSfasta ->
      let aa,_ = read_fasta fname in
      {seq with ss=(ss2idx aa.(0)); seqlen=String.length aa.(0);}
  | SASA ->
      let floats = read_floats fname in
      {seq with sasa=floats; seqlen=Array.length floats;}
  | Index ->
      let index = read_ints fname in
      Hashtbl.add seq.indexes !idxname index;
      seq
  | Index1 ->
      let index = read_ints ~skip1:true fname in
      Hashtbl.add seq.indexes !idxname index;
      seq
  | Threshold ->
      let floats = read_floats fname in
      let index = Array.map (fun x-> if x<thresh then 0 else 1) floats in
      Hashtbl.add seq.indexes !idxname index;
      seq
  | TClass ->
      let index = read_ints fname in
      let m = Array.length index in
      let n = 1 + (Array.fold_left max 0 index) in
      let tclass = Array.make_matrix m n 0.0 in
      for i=0 to m-1 do
	tclass.(i).(index.(i)) <- 1.0;
      done;
      {seq with tclass=tclass; seqlen=Array.length tclass;}
  | P2Class ->
      let floats = read_floats fname in
      let m = Array.length floats in
      let pclass = Array.make_matrix m 2 0.0 in
      for i=0 to m-1 do
	pclass.(i).(0) <- 1.0 -. floats.(i);
	pclass.(i).(1) <- floats.(i);
      done;
      {seq with pclass=pclass; seqlen=Array.length pclass;}
  | Floats ->
      let floats = read_floats fname in
      let m = Array.length floats in
      {seq with floats=floats; seqlen=m;}
  end
  with Scan_failure s -> raise (Scan_failure (fname^":"^s))

    
exception Bad_sequence of string * string * int * int

let feature_lengths s = [
    "Sequence",  Array.length s.seq;
    "Profile",   s.prof.len;
    "SS",        Array.length s.ss;
    "SSpred",    Array.length s.sspred;
    "SSprof",    Array.length s.ssprof;
    "SASA",      Array.length s.sasa;
    "TClass",    Array.length s.tclass;
    "PClass",    Array.length s.pclass;
    "Floats",    Array.length s.floats;
  ];;

(** Sanity checks for sequence properties.   *)
let verify_seq id s =
  let checks = feature_lengths s in
  let op (str,n) =
    if n>0 && n <> s.seqlen then
      raise (Bad_sequence (id,str,n,s.seqlen))
  in
  List.iter op checks;
  let check_index idxname index =
    let n = Array.length index in
    if n <> s.seqlen then
      raise (Bad_sequence (id,idxname,n,s.seqlen));
  in
  Hashtbl.iter check_index s.indexes


(*   for i=0 to (Array.length s.indexes)-1 do *)
(*     let n = Array.length s.indexes.(i) in *)
(*     if n <> s.seqlen then *)
(*       raise (Bad_sequence (id,"Index"^(string_of_int i),n,s.seqlen)); *)
(*   done;; *)

