open Printf;;
open Scanf;;

(** The essential identity! *)
let ident x = x;;

(** Floatin point maximum *)
let fmax : (float -> float -> float) = max

let fabs = Pervasives.abs_float

(** Reverse comparator *)
let revcomp x y = -(compare x y);;

(** Convert a character to a length 1 string *)
let ch2str c = let s= String.create 1 in s.[0]<-c; s;;

(** Convert a list of characters to a string *)
let chars2string l =
  let n = List.length l in
  let s = String.create n in
  let ptr = ref l in
  for i=0 to n-1 do
    s.[i]<-List.hd !ptr; ptr:=List.tl !ptr;
  done;
  s;;

(** Reverse the contents of an array *)
let reverse_array a =
  let n = Array.length a in
  for i=0 to n/2-1 do
    let x = a.(i) in
    a.(i) <- a.(n-i-1);
    a.(n-i-1) <- x;
  done;;

(** Convert a list to an array then reverse it *)
let revlist2array l =
  let a = Array.of_list l in
  reverse_array a;
  a;;

(** Create an array from a queue *)
let q2array q =
  let n = Queue.length q in
  if n=0 then [||]
  else begin
    let a = Array.make n (Queue.peek q) in
    let i = ref 0 in
    let iter x = a.(!i) <- x; incr i in
    Queue.iter iter q;
    a
  end;;

(** Create a list from a queue *)
let q2list q =
  List.rev (Queue.fold (fun l x->x::l) [] q);;

(** Fill a matrix [a] with value [v] *)
let fill_matrix a v =
  let n = Array.length a.(0) in
  Array.iter (fun row-> Array.fill row 0 n v) a;;


(** Ensure size of a matrix is at least size [m] by [n] *)
let check_matsize m n a =
  Array.length a >= m && Array.length a.(0) >= n;;


(** Simple linked list implementation *)
module Llist = struct
  type 'a node = {
    dat:'a; 
    mutable prev:'a node option; 
    mutable next:'a node option; 
  }
  type 'a llist = {
    mutable nnodes: int; 
    mutable head:'a node option; 
    mutable tail:'a node option
  }
  let empty () = {nnodes=0; head=None; tail=None}
  let isempty l = l.nnodes = 0
  let append l v = 
    let vn = {dat=v; prev=l.tail; next=None} in
    l.nnodes <- l.nnodes+1;
    match l.tail with
      None -> l.head <- Some(vn); l.tail <- Some(vn)
    | Some(tl) -> tl.next <- Some(vn); l.tail <- Some(vn)
  let to_list l =
    let rec f nopt = match nopt with
	None -> []
      | Some(nd) -> nd.dat::(f nd.next) in
    f l.head
  let to_array l =
    let ptr = ref l.head in
    let f _ = match !ptr with Some(nd)-> ptr:=nd.next; nd.dat; | None -> failwith "Problem" in
    Array.init l.nnodes f
  let length l = l.nnodes
  let top l = match l.head with Some(v)->v.dat | None -> failwith "Empty Llist"
  let single v = 
    let n = {dat=v; prev=None; next=None} in 
    {nnodes=1; head=Some(n); tail=Some(n)}
  let append_list ll l = List.iter (fun x->append ll x) l
end    
  
(** Open file [fname] and read its content: whitespace separated
    strings *)
let read_strings fname = 
  let b = Scanning.from_file fname in
  let q = Queue.create () in
  while not (Scanning.end_of_input b) do
    let s = bscanf b " %s " ident in
    Queue.add s q;
  done;
  q2list q;;

(** Parse out strings in [s] and return them as an array *)
let parse_strings s = 
  let b = Scanning.from_string s in
  let q = Queue.create () in
  while not (Scanning.end_of_input b) do
    let s = bscanf b " %s " ident in
    Queue.add s q;
  done;
  q2array q;;


(** Scan the number of lines in the file, done as efficiently as I can
    figure out how *)
let count_lines file = 
  let f = open_in file in
  let n = ref 0 in
  begin try while true do
    if input_char f = '\n' then incr n else ();
  done;
  with End_of_file -> ()  end;
  close_in f;
  !n

(** Count the number of fields inth e first line of a file *)
let count_fields file =
  let f = open_in file in
  let l = input_line f in
  let b = Scanning.from_string l in
  close_in f;
  let n = ref 0 in
  begin try while not (Scanning.end_of_input b) do 
    ignore(bscanf b " %s " ident); 
    incr n;
  done with End_of_file -> () end;
  !n

(** Tricky implementation of quicksort.  Supports optional size
    argument [stop] which defaults to size of the array. *)
let rec quicksort ?(start=0) ?stop a =
  let stop = match stop with None -> (Array.length a)-1 | Some(s) -> s-1 in
  if start < stop then begin
    let x = a.(stop) in
    let i = ref (start-1) in
    for j=start to stop-1 do
      if a.(j) < x then begin
	incr i;
	let t = a.(!i) in
	a.(!i) <- a.(j);
	a.(j) <- t
      end else ();
    done;
    incr i;
    a.(stop) <- a.(!i);
    a.(!i) <- x;
    quicksort ~start ~stop:(!i) a;
    quicksort ~start:(!i+1) ~stop:(stop+1) a
  end else ()

(** Sort array a and return an ordering to sort other arrays. *)
let order a =
  let n = (Array.length a) in
  let b = Array.init n ident in
  let rec quicksort start stop  =
    if start < stop then begin
      let x,xi = a.(stop),b.(stop) in
      let i = ref (start-1) in
      for j=start to stop-1 do
	if a.(j) < x then begin
	  incr i;
	  let t, ti = a.(!i), b.(!i) in
	  a.(!i) <- a.(j); b.(!i) <- b.(j);
	  a.(j) <- t;      b.(j) <- ti;
	end else ();
      done;
      incr i;
      a.(stop) <- a.(!i);  b.(stop) <- b.(!i);
      a.(!i) <- x;         b.(!i) <- xi;
      quicksort start (!i-1);
      quicksort (!i+1) (stop);
    end else ()
  in
  quicksort 0 (n-1);
  b

(** Read a matrix from file name [fname], assumed floating point  *)
let read_matrix fname =
  let m = count_lines fname in
  let n = count_fields fname in
  let a = Array.make_matrix m n 0.0 in
  let b = Scanning.from_file fname in
  for i=0 to m-1 do 
    for j=0 to n-1 do
      a.(i).(j)<-bscanf b " %f " ident
    done; 
  done;
  a

(** Read integers from a file, optionally skip first line *)
let read_ints ?(skip1=false) fname =
  let f = Scanning.from_file fname in
  if skip1 then	bscanf f "%s@\n" ignore;
  let ints = Queue.create () in
  while not (Scanning.end_of_input f) do
    let x = bscanf f " %d " ident in
    Queue.add x ints;
  done;
  q2array ints;;

(** Read floats from a file, optionally skip first line *)
let read_floats ?(skip1=false) fname =
  let f = Scanning.from_file fname in
  if skip1 then	bscanf f "%s@\n" ignore;
  let floats = Queue.create () in
  while not (Scanning.end_of_input f) do
    let x = bscanf f " %f " ident in
    Queue.add x floats;
  done;
  q2array floats;;

(** Dot product of two vectors, no checks for equal length *)
let dotp x y = 
  let width = Array.length x in
  let s = ref 0.0 in
  for i=0 to width-1 do s:= !s +. x.(i)*.y.(i); done;
  !s;;

(** L2-norm of a vector *)
let norm2 x = sqrt (dotp x x)

(** Calculates the vector [mat * x] *)
let matrix_vector_prod mat x =
  let n = Array.length x in
  assert(Array.length mat = n);
  let y = Array.make n 0.0 in
  for i=0 to n-1 do
    y.(i) <- dotp mat.(i) x;
  done;
  y;;

(** Calculates [x' * mat * y] using the equality to 
    [sum(sum( mat .* (x*y')))] *)
let inner_product mat x y =
  let n = Array.length x in
  assert(Array.length mat = n && Array.length mat.(0) = n && Array.length y = n);
  let sum = ref 0.0 in
  for i=0 to n-1 do
    for j=0 to n-1 do
      sum := !sum +. x.(i) *. mat.(i).(j) *. y.(j);
    done;
  done;
  !sum;;

(** Sum an array of floating point numbers *)
let array_sumf = Array.fold_left (+.) 0.0;;
(* let array_sumf ~n a = *)
(*   let n = match n with  *)
(*       None -> Array.length a  *)
(*     | Some(n) -> n *)
(*   in *)
(*   let sum = ref 0.0 in *)
(*   for i=0 to n-1 do sum := !sum +. a.(i); done; *)
(*   !sum *)

(** Write a matrix to a file with each entry writting according to
    [form]. If no file name is provided, write it to standard
    output *)
let write_matrix ?fname ?m ?n form a =
  let f = match fname with
      Some(fname) -> open_out fname
    | None -> stdout in
  let m = match m with
      Some(m) -> m
    | None -> Array.length a in
  let n = match n with 
      Some(n) -> n
    | None -> Array.length a.(0) in
  for i=0 to m-1 do 
    for j=0 to n-1 do
      fprintf f form a.(i).(j)
    done;
    fprintf f "\n"
  done;
  match fname with
    Some(_) -> close_out f
  | None -> ();;

(** Create a new string from [s] with line breaks. [w] controls about
    how wide each line should be. [hard] specifies whether the breaks
    are hard, if not a word break is searched for and chosen as the
    break. *)
let wordwrap ?(hard=false) ?(sep="\n") w s =
  let n = String.length s in
  let sl = ref [] in
  let idx = ref 0 in
  while !idx < n do
    if !idx+w >= n then begin
      sl := (String.sub s !idx (n - !idx))::!sl;
      idx := !idx + w;
    end
    else if hard then begin
      sl := (String.sub s !idx w)::!sl;
      idx := !idx + w;
    end
    else begin
(*       let stop = ref (!idx+w) in *)
      let stop = ref (!idx) in
      while !stop < !idx+w && s.[!stop] <> '\n' do incr stop; done;
      if !stop == !idx+w then
	while s.[!stop]<>' ' do decr stop done;
      sl := (String.sub s !idx (!stop - !idx))::!sl;
      idx := !stop+1;
    end
  done;
  String.concat sep (List.rev !sl);;

(** Return index of substring s in string t *)
let find_substr s t =
  let m = String.length s in
  let n = String.length t in
  let rec check_chars i j =
    if i<m && j=n then false
    else if i=m && j<=n then true
    else if s.[i] <> t.[j] then false
    else check_chars (i+1) (j+1) in
  let rec search start c =
    let j = try 
      String.index_from t start c
    with _ -> raise Not_found in (*  (sprintf "find_substr: %s not found in %s" s t)) in *)
    if check_chars 0 j then j else search (j+1) c in
  search 0 s.[0];;

let contains_substr s t = 
  try 
    ignore(find_substr s t);
    true;
  with Not_found -> false;;

(** Return index of subarray s in array t *)
let find_subarray s t =
  let m = Array.length s in
  let n = Array.length t in
  let rec check i j =
    if i<m && j=n then false
    else if i=m && j<=n then true
    else if s.(i) <> t.(j) then false
    else check (i+1) (j+1) in
  let rec search start c =
    let j = try
      let s = ref start in
      while t.(!s) <> c do incr s; done;
      !s
    with _ -> raise Not_found in (*  (sprintf "find_substr: %s not found in %s" s t)) in *)
    if check 0 j then j else search (j+1) c in
  search 0 s.(0);;

(** Eliminate directory and extension from filename *)
let basename f =
  Filename.basename 
    (try Filename.chop_extension f
     with Invalid_argument _ -> f);;

(** Simple timing functions, allows an interval to be kept by calling [interval ()] *)
module Timing = struct
  let last = ref 0.0 
  let interval () =
    let t0 = !last in
    let t = Sys.time () in
    last := t;
    t -. t0;
end

(** Open output file if name given, otherwise standard out *)
let opt_open fout =  
  match fout with 
    None -> stdout
  | Some(s) -> open_out s;;

(** Close output if name given, otherwise nothing *)
let opt_close fout f =
  match fout with
    None -> () 
  | Some(_) -> close_out f;;

(** Create an array of pairs of elements from [a] and [b] *)
let pair_arrays a b =
  Array.mapi (fun i ax -> (ax, b.(i))) a;;


(** Give an array of elements from [a] that satisfy the filter [f] *)
let arrayfilt f a =
  let l = ref [] in
  Array.iteri (fun i x -> if f x then l:=i::!l) a;
  let l = ref (List.rev !l) in
  Array.init (List.length !l) (fun _->let i=List.hd !l in l:=List.tl !l; a.(i));;

(** Return the index of the first array element satisfy predicate [f] *)
let array_first f a =
  let n = Array.length a in
  let rec g i =
    if i < n then
      if f a.(i) then i
      else g (i+1)
    else raise Not_found
  in g 0
      
(** Vector add a*x to y *)
let axpy a x y =
  let n = Array.length x in
  for i=0 to n-1 do
    y.(i) <- y.(i) +. a *. x.(i)
  done


(** Search for a file in the given list of directories.  Searches
    current directory if list is empty.  Return file name if found
    or raise an exception if not found. *)
module Find = struct

  exception File_not_found of string * string
  let dirs = 
    let q = Queue.create () in
    Queue.add "" q;
    q

  let dirstr = ref "."

  let add_dir d = 
    Queue.add (d^"/") dirs;
    dirstr := sprintf "%s:%s" !dirstr d
    
  let print_dirs sep chan = 
    Queue.iter (fun d-> fprintf chan "%s%s" d sep) dirs

  let file fname =
    let fn o d = match o with
	Some(_) -> o
      | None ->
	  let f = sprintf "%s%s" d fname in
	  if Sys.file_exists f then Some(f)
	  else None
    in
    match Queue.fold fn None dirs with
      Some(f) -> f
    | None -> raise (File_not_found (fname, !dirstr))
  
end


(** Transpose a matrix *)
let transpose x =
  let m,n = (Array.length x), (Array.length x.(0)) in
  let y = Array.make_matrix n m x.(0).(0) in
  for i=0 to m-1 do
    for j=0 to n-1 do
      y.(j).(i) <- x.(i).(j);
    done;
  done;
  y

(** Add to a queue, reverse arguments for ease of mapping *)
let queue_add q k = Queue.add k q

(** Read float, handle nans and inf if needed *)
let rfloat f = bscanf f " %s " (fun s-> float_of_string s)

let rnan = rfloat


let isnan f = (classify_float f) = FP_nan

(** Simple statistics calculations *)
module Avgstd = struct
  type t = {
    mutable cnt:float;
    mutable wgt:float;
    mutable avg:float;
    mutable std:float;
  }

  let make () = {cnt=0.0; wgt=0.0; avg=0.0; std=0.0 }

  let update s ?(w=1.0) x = 
    s.cnt <- s.cnt +. 1.0;
    s.wgt <- s.wgt +. w;
    let del = x -. s.avg in
    s.avg <- s.avg +. (w /. s.wgt) *. del;
    s.std <- s.std +. w *. del *. (x -. s.avg)

  let update_ignore_nan s ?(w=1.0) x =
    if not (isnan x) then update s ~w x

  let finalize s =
    s.avg, (sqrt (s.cnt /. (s.cnt -. 1.0) *. (s.std /. s.wgt))), s.cnt

(*   let update s ?(ignore_nan=false) ?(w=1.0) x =  *)
(*     if not (ignore_nan && isnan x) then begin *)
(*       s.cnt <- s.cnt +. 1.0; *)
(*       s.wgt <- s.wgt +. w; *)
(*       let del = x -. s.avg in *)
(*       s.avg <- s.avg +. (w /. s.wgt) *. del; *)
(*       s.std <- s.std +. w *. del *. (x -. s.avg) *)
(*     end *)

end

module Cor = struct
  type t = {
    mutable cnt:float;
    mutable real_cnt:float;
    mutable xsum:float;
    mutable ysum:float;
    mutable x2sum:float;
    mutable y2sum:float;
    mutable xysum:float;
    mutable sqdif:float;
    xstats:Avgstd.t;
    ystats:Avgstd.t;
  }
  let make () = {
    cnt=0.0; real_cnt=0.0;
    xsum=0.0; ysum=0.0; x2sum=0.0; y2sum=0.0; xysum=0.0; sqdif=0.0; 
    xstats = Avgstd.make (); ystats = Avgstd.make ();
  }
    
  let update s x y =
    s.cnt <- s.cnt +. 1.0;
    s.real_cnt <- s.real_cnt +. 1.0;
    s.xsum <- s.xsum +. x;
    s.ysum <- s.ysum +. y;
    s.x2sum <- s.x2sum +. x *. x;
    s.y2sum <- s.y2sum +. y *. y;
    s.xysum <- s.xysum +. x *. y;
    let dif = x -. y in
    s.sqdif <- s.sqdif +. dif *. dif;
    Avgstd.update s.xstats x;
    Avgstd.update s.ystats y

  let update_ignore_nan s x y =
    if not (isnan x) && not (isnan y) 
    then update s x y
    else s.real_cnt <- s.real_cnt +. 1.0

  let cor s =
    let num = s.xysum -. (s.xsum *. s.ysum /. s.cnt) in
    let denx = s.x2sum -. (s.xsum *. s.xsum /. s.cnt) in
    let deny = s.y2sum -. (s.ysum *. s.ysum /. s.cnt) in
    let cor = num /. (sqrt (denx *. deny)) in
    cor
      
  let res2 s     = s.sqdif			    (* Squared residual *)
  let rmsd s     = sqrt (s.sqdif /. s.cnt)	    (* Root mean squared deviation *)
  let res2norm s = (sqrt s.sqdif) /. (sqrt s.x2sum) (* Normalized mean square error *)
  let mse s      = (sqrt s.sqdif) /. s.cnt	    (* Mean squared error *)
  let cov s      = s.cnt /. s.real_cnt		    (* Coverage of non-nan predictions *)

  let x2sum s = s.x2sum
  let y2sum s = s.y2sum

  let get_xstats s = Avgstd.finalize s.xstats 
  let get_ystats s = Avgstd.finalize s.ystats 

(*   let finalize s = *)
(*     let corr = cor s in *)
(*     let res = res2 s in *)
(*     let xs = get_xstats s in *)
(*     let ys = get_ystats s in *)
(*     corr, res, xs, ys *)

end

module Err = struct
  type t = 
      NoCheck				(**No checking*)
    | Complain 				(**Issue Warnings*)
    | Panic;;				(**Bail out if problems happen*)
(*     | Whine				(\**Issue more warnings *\) *)

  (** Global error checking level *)
  let lvl = ref NoCheck;;
  let lvls = ["none";"complain";"panic"];;

  (** Documentation strings for error levels *)
  let docs = [
    "none","Don't issue any warnings";
    "complain","Issue a warnings, usually a lot, for things that look wrong";
    "panic","Bail out if something looks bad";
  ]

  (** Set the error checking based on argument *)
  let setlvl s = 
    lvl := match s with
      "none" -> NoCheck
    | "complain" -> Complain
    | "panic" -> Panic
    | _ -> failwith ("Unkown error level " ^ s);;

  let warn msg = eprintf "Warning: %s\n" msg

  let decide msg def = 
    match !lvl with
      NoCheck -> def
    | Complain -> eprintf "Warning: %s\n" msg; def
    | Panic -> failwith msg
end

(** If [tf] is true, open the given filename.  If false, open a null
    file for output.*)
let cond_out tf fname = open_out (if tf then fname else "/dev/null")

module Queue_hash = struct
  type ('a , 'b) t = ('a , ('b Queue.t)) Hashtbl.t
  let add h k v =
    match Hashtbl.mem h k with
      true -> let q = Hashtbl.find h k in Queue.add v q
    | false ->
	let q = Queue.create () in
	Queue.add v q;
	Hashtbl.add h k q

  let create = (Hashtbl.create : int -> ('a , 'b) t)
  let find = Hashtbl.find
  let mem = Hashtbl.mem
  let iter = Hashtbl.iter
  let exportl qh =
    let f k v acum = (k, q2array v)::acum in
    Hashtbl.fold f qh []
  let exporta qh =
    Array.of_list (exportl qh)
end
      
(** Get all keys for the given hash table [h] *)
let hash_keys h =
  Hashtbl.fold (fun k _ l-> k::l) h []

(** Permute the entries of array [a], go through it [m] times *)
let permute m a = 
  let n = Array.length a in
  for ii=1 to m do
    for i=0 to n-1 do
      let j = Random.int n in
      let x = a.(i) in
      a.(i) <- a.(j);
      a.(j) <- x;
    done;
  done
