(** Module to do weighting of alignment neighbors *)

open Printf
open Sequence
open Query_align

let local_width = ref 2

(** Shift [scores] up to [newmin] *)
let shift_to_min newmin scores =
  let minscr = Array.fold_left min min_float scores in
  let shift = if minscr < newmin then newmin -. minscr else 0.0 in
  for i=0 to (Array.length scores)-1 do
    scores.(i) <- scores.(i) +. shift
  done;;

(** Divide each element of [scores] by the sum of [scores] *)
let normalize scores = 
  let sum = Util.array_sumf scores in
  for i=0 to (Array.length scores)-1 do
    scores.(i) <- scores.(i) /. sum
  done;;

(** Equal weighting scheme *)
let equal_weights wgts _ _ _ _ =
  for i=0 to (Array.length wgts)-1 do
    for j=0 to (Array.length wgts.(i))-1 do
      wgts.(i).(j) <- 1.0;
    done;
  done

(** Weights are the same for each sequence, based on original
    alignment score. Scores are shifted to one if any negative scores
    are present and normalized to get the weights. Note that the
    'matrix' returned should *not* be modified as the row pointers all
    point to the same row *)
let global_weights wgts _ qseq best _ =
  let m, n = qseq.seqlen, Array.length best in
  let scores = Array.map (fun (s,_)->s) best in
  shift_to_min 1.0 scores;
  for i=0 to m-1 do
    for j=0 to n-1 do
      wgts.(i).(j) <- scores.(j)
    done;
  done;;
  
(** Weights vary by position and sequence. A local sequence alignment
    score is computed, shifted, and normalized to determine the
    weight. [width] is the number *alignment* of positions around the
    central residue for the computation. *)
let local_weights wgts scoremat qseq best seqDB =
  let width = !local_width in
  Util.fill_matrix wgts 0.0;
  let nbest = Array.length best in
  for i=0 to nbest-1 do
    let _, hit = best.(i) in
    let dbseq = Hashtbl.find seqDB hit.dbkey in
    Scoring.setup_scoring_mat ~reset:true scoremat qseq dbseq;
(*     Util.write_matrix ~fname:("score."^(string_of_int i)^"mat") "%8.2f " scoremat; *)
    for j=0 to hit.alen-1 do
      match hit.qal.(j) with
	Some(qdx) ->
	  let start = max (j-width) 0 in
	  let stop = min (j+width) hit.alen in
	  let s = Align.score_alignment hit.qal hit.dal scoremat start stop in
(* 	  printf "%d %d %d %d %d %f\n" j qdx start stop len s; *)
	  wgts.(qdx).(i) <- s
      | None -> ()
    done;
  done;
(*   Util.write_matrix ~fname:("X.wgt") ~m:(qseq.seqlen) ~n:(nbest) "%f " wgts; *)
  Array.iter (shift_to_min 1.0) wgts


open Ckarg

let scheme = ref equal_weights

(** Options controlling module behavior  *)
let options = [
  "-wgt_scheme","key","Weighting scheme for topk matches.",
    SymbolArb(true,[
      "equal","Each topk hit is weighted equally",(fun ()-> scheme := equal_weights);
      "global","Hits are weighted by how well they score against the query, \
         scores taken from the original alignment",(fun ()-> scheme:= global_weights);
      "local","Weights depend on alignment score in a window around query/DB",
        (fun ()-> scheme:= local_weights);
    ]);
  "-local_width","int","Width of window for local weight determination",
    Set_int(local_width);
]  @ Scoring.options @ Align.options


